# -*- coding: utf-8 -*-
"""
Implementation of 'AmqpConfigurationManager' class
"""
from __future__ import absolute_import

# Import python libs
import codecs
import os
import os.path
import json

# Import local libs
# pylint: disable=import-error
import niskyline.messagebus.exceptions as exceptions
from niskyline.messagebus.amqp_configuration import (
    AmqpConfiguration, SKYLINE_LOCALHOST_CONFIGURATION_ID,
    SKYLINE_MASTER_CONFIGURATION_ID
)
from niskyline.messagebus.paths import get_skyline_configurations_directory
# pylint: enable=import-error


class AmqpConfigurationManager(object):
    """
    Factory for AMQPConfiguration objects.
    """
    _instance = None

    def __init__(self, id_=None, enable_fallbacks=True):
        """
        :param id_: The configuration ID to use or ``None`` to use
            the default configuration.
        :type id_: str or None
        :param enable_fallbacks: If ``True`` and the desired configuration
            ID is not found, will fall back to the default configuration.
            If ``False`` and the desired configuration is not found, will
            raise a :class:`niskyline.messagebus.exceptions.SkylineException`.
        :type enable_fallbacks: bool
        """
        self._closing = False
        self._configurations = {}
        self._read_file_set = set()
        self._configurations = self._read_configurations()
        self._default_configuration = self._get_configuration_helper(
            id_, enable_fallbacks, False
        )

    def __del__(self):
        self.close()

    def close(self):
        """
        Close this instance of :class:`AmqpConfigurationManager`.
        """
        if self._closing:
            return
        self._closing = True
        cls = self.__class__
        if cls._instance == self:  # pylint: disable=protected-access
            cls._instance = None  # pylint: disable=protected-access

    @classmethod
    def shutdown(cls):
        """
        If a singleton instance of :class:`AmqpConfigurationManager` exists,
        close and remove it.
        """
        if cls._instance:
            cls._instance.close()
            cls._instance = None

    @classmethod
    def get_instance(cls):
        """
        Obtain a singleton instance of :class:`AmqpConfigurationManager`.
        """
        if cls._instance is None:
            cls._instance = cls(None, True)
        return cls._instance

    @classmethod
    def get_configuration(cls, id_=None, enable_fallbacks=True, refresh=True):
        """
        Obtain an instance of
        :class:`niskyline.messagebus.amqp_configuration.AmqpConfiguration`
        from the singleton instance of :class:`AmqpConfigurationManager`.

        :param id_: The configuration ID to use or ``None`` to use
            the default configuration.
        :type id_: str or None
        :param enable_fallbacks: If ``True`` and the desired configuration
            ID is not found, will fall back to the default configuration.
            If ``False`` and the desired configuration is not found, will
            raise a :class:`niskyline.messagebus.exceptions.SkylineException`.
        :type enable_fallbacks: bool
        :param refresh: If ``True``, will refresh the selected configuration
            if it has changed since it was last loaded. If ``False``, will not
            refresh and could potentially use stale data.
        :type refresh: bool
        """
        instance = cls.get_instance()
        if id_ is None:
            if refresh:
                instance._default_configuration.refresh()  # pylint: disable=protected-access
            return instance._default_configuration  # pylint: disable=protected-access
        return instance._get_configuration_helper(id_, enable_fallbacks, refresh)  # pylint: disable=protected-access

    def _get_configuration_helper(self, id_, enable_fallbacks, refresh):
        """
        Helper function to obtain an instance of
        :class:`niskyline.messagebus.amqp_configuration.AmqpConfiguration`.

        :param id_: The configuration ID to use or ``None`` to use
            the default configuration.
        :type id_: str or None
        :param enable_fallbacks: If ``True`` and the desired configuration
            ID is not found, will fall back to the default configuration.
            If ``False`` and the desired configuration is not found, will
            raise a :class:`niskyline.messagebus.exceptions.SkylineException`.
        :type enable_fallbacks: bool
        :param refresh: If ``True``, will refresh the selected configuration
            if it has changed since it was last loaded. If ``False``, will not
            refresh and could potentially use stale data.
        :type refresh: bool
        """
        if refresh:
            # Check for new files
            self._configurations.update(
                self._read_configurations(only_new_files=True)
            )

        selected_amqp_configuration = None
        if not id_:
            selected_amqp_configuration = self._fallback()
            if selected_amqp_configuration is None:
                raise exceptions.SkylineException(
                    'No Skyline configurations available.',
                    exceptions.ERROR_NO_SKYLINE_CONFIGURATIONS_AVAILABLE
                )
        else:
            selected_amqp_configuration = self._configurations.get(id_)
            if selected_amqp_configuration is None:
                if enable_fallbacks:
                    selected_amqp_configuration = self._fallback()
                    if selected_amqp_configuration is None:
                        msg = 'No Skyline configuration \'' + id_ + \
                                '\' found, and no fallback configurations available.'
                        raise exceptions.SkylineException(
                            msg,
                            exceptions.ERROR_NO_SKYLINE_CONFIGURATIONS_AVAILABLE
                        )
                else:
                    msg = 'No Skyline configuration \'' + id_ + \
                            '\' found, and fallback configurations were not enabled.'
                    raise exceptions.SkylineException(
                        msg,
                        exceptions.ERROR_NO_SKYLINE_CONFIGURATIONS_AVAILABLE
                    )
        if refresh:
            selected_amqp_configuration.refresh()
        return selected_amqp_configuration

    def _read_config_file(self, file_path):
        """
        Read the configuration file.

        :param file_path: Path to the configuration file.
        :type file_path: str
        :return: A tuple containging the configuration ID and a
             :class:`niskyline.messagebus.amqp_configuration.AmqpConfiguration`
             object containing the configuration information.
        :rtype: tuple(str, niskyline.messagebus.amqp_configuration.AmqpConfiguration)
            or tuple(None, None)
        """
        last_modifed_time = os.path.getmtime(file_path)
        with codecs.open(file_path, 'r', encoding='utf-8-sig') as fp_:
            configuration = json.loads(fp_.read())

        if 'Id' in configuration:
            configuration_id = configuration['Id'].lower()
            if configuration_id in self._configurations:
                msg = ('Duplicate ID "{0}" in Skyline configuration file '
                       '"{1}"'.format(configuration_id, file_path))
                raise exceptions.SkylineException(
                    msg,
                    exceptions.ERROR_DUPLICATE_CONFIGURATION_ID
                )
            amqp_config = AmqpConfiguration(
                file_path, configuration, last_modifed_time
            )
            return configuration_id, amqp_config
        return None, None

    def _read_configurations(self, only_new_files=False):
        """
        Read configurations by reading all '.json' files in the
        configuration directory.

        :param only_new_files: If ``True``, read only files not read since
            last scan.
        :type only_new_files: bool
        :return: A dictionary with key of the configuration ID and value
             an instance of
             :class:`niskyline.messagebus.amqp_configuration.AmqpConfiguration`.
        :rtype: dict(str, niskyline.messagebus.amqp_configuration.AmqpConfiguration)
        """
        configurations = {}
        path = get_skyline_configurations_directory()
        only_files = [file_name for file_name in os.listdir(path)
                      if os.path.isfile(os.path.join(path, file_name))]
        for file_name in only_files:
            file_path = os.path.join(
                path,
                file_name
            )
            if file_name.endswith('.json'):
                if not only_new_files or file_path not in self._read_file_set:
                    # Read .json configuration file
                    configuration_id, amqp_config = self._read_config_file(file_path)
                    if configuration_id is not None:
                        configurations[configuration_id] = amqp_config
                        self._read_file_set.add(file_path)

        return configurations

    def _fallback(self):
        """
        Helper function to facilitate falling back on the default configuration
        ID when the desired ID is not available.
        """
        if SKYLINE_LOCALHOST_CONFIGURATION_ID in self._configurations:
            return self._configurations[SKYLINE_LOCALHOST_CONFIGURATION_ID]
        return self._configurations.get(
            SKYLINE_MASTER_CONFIGURATION_ID,
            None
        )
