# -*- coding: utf-8 -*-
"""
Implementation of 'Error' class
"""
from __future__ import absolute_import

# Import python libs
import json

# Import local libs
# pylint: disable=import-error
import niskyline.messagebus.exceptions as exceptions
# pylint: enable=import-error


class Error(object):  # pylint: disable=too-many-public-methods
    """
    Skyline Error information.

    Used to encapsulate error information when an error occurs
    in the Skyline framework.
    """
    def __init__(self, error_code=None, info=None, args=None):
        """
        :param error_code: The error code of the error or ``None`` if no
            error has occurred.
        :type error_code: niskyline.messagebus.error_code.ErrorCode or None
        :param info: Error information string. May be ``None`` in the cases
            of no error or if this information is desired to be auto-generated
            based on ``error_code``.
        :type info: str or None
        :param args: A list of arguments associated with the error.
        :type args: list
        """
        self._has_value = False
        self._error_code = None
        self._info = None
        self._args = None
        self.set(error_code, info, args)

    def set(self, error_code, info, args):
        """
        Set an error.

        :param error_code: The error code of the error or ``None`` if no
            error has occurred.
        :type error_code: niskyline.messagebus.error_code.ErrorCode or None
        :param info: Error information string. May be ``None`` in the cases
            of no error or if this information is desired to be auto-generated
            based on ``error_code``.
        :type info: str or None
        :param args: A list of arguments associated with the error.
        :type args: list
        """
        self._error_code = error_code
        self._info = info
        if self._info is None:
            self._info = ''
        self._args = args
        if self._args is None:
            self._args = []
        self._has_value = bool(error_code is not None)
        if self._has_value and self._error_code._message and info is None:  # pylint: disable=protected-access
            self._info = self._format_message()

    def has_value(self):
        """
        Return whether this :class:`Error` object has a value.

        :return: ``True`` if this :class:`Error` object has a value. ``False``
            otherwise.
        :rtype: bool
        """
        return self._has_value

    def reset(self):
        """
        Reset this :class:`Error` object to a no error state.
        """
        self._has_value = False
        self._error_code = None
        self._info = ''
        self._args = []

    def serialize(self):
        """
        Return a JSON representation of the error string.

        :return: A JSON representation of the error string.
        :rtype: str
        """
        if self._has_value:
            return '\"' + json.loads(self.to_string()) + '\"'
        return 'null'

    def get_name(self):
        """
        Get the error name information.

        :return: The error name information or ``None`` if no error is set.
        :rtype: str or None
        """
        if not self._has_value:
            return None
        return self._error_code.to_string()

    def get_code(self):
        """
        Get the error code.

        :return: The error code or ``None`` if no error is set.
        :rtype: niskyline.messagebus.error_code.ErrorCode or None
        """
        return self._error_code

    def get_info(self):
        """
        Get the error information.

        :return: The error information. Will return an
            empty string if there is no associated error information.
        :rtype: str
        """
        return self._info

    def set_info(self, info):
        """
        Set the error information.

        :param info: The error information.
        :type info: str
        """
        self._info = info
        if self._info is None:
            self._info = ''

    def get_args(self):
        """
        Get the arguments.

        :return: The arguments. Will return an
            empty list if there is no associated arguments.
        :rtype: list
        """
        return self._args

    def set_args(self, args):
        """
        Set the arguments.

        :param args: The arguments.
        :type args: list
        """
        self._args = args
        if self._args is None:
            self._args = []

    def to_string(self):
        """
        Return a string representation of the error.

        :return: A string representation of the error.
        :rtype: str
        """
        if not self._has_value:
            return ''
        code = self._error_code.to_string()
        if not self._info:
            return code
        return code + ': ' + self._info

    def _format_message(self):  # pylint: disable=too-many-branches
        """
        This method performs message inserts on the ErrorCode's message string
        and returns the result. The message format follows the C#
        string.Format(<message>, params ...) syntax.  Message insert arguments
        are denoted by braces enclosing an integer value.  ex:  '{#}' For
        example: "Processing file {0} resulted in {1} failed operation(s)."
        The vector of string arguments included in the Error class is used as
        the positional inserts for the message.

        :return: The formatted message.
        :rtype: str
        """
        tokens = []
        token = ''
        for char in self._error_code._message:  # pylint: disable=protected-access
            if char == '{':
                # Found a start delimiter. Push the current token and start a new token.
                if token != '':
                    tokens.append(token)
                    token = ''

            # Append the current character to the token
            token += char

            if char == '}':
                # Found an end delimiter.  Push the current token and start a new token.
                if token != '':
                    tokens.append(token)
                    token = ''

        # Handle the last token
        if token != '':
            tokens.append(token)

        formatted_message = ''
        for token in tokens:
            if len(token) > 2:
                if token[0] == '{' and token[len(token)-1] == '}':
                    # This is an insert token
                    # Get the message insert number from the string, and convert it to an integer
                    insert_number_text = token[1:len(token)-2]
                    try:
                        insert_number = int(insert_number_text)
                    except ValueError:
                        raise exceptions.SkylineException(
                            'Malformed message insert number.'
                        )

                    if not len(self._args) > insert_number:
                        raise exceptions.SkylineException(
                            'Too few inserts passed to format the message.'
                        )

                    formatted_message += self._args[insert_number]
                else:
                    # Not an insert.  Append the string to the message
                    formatted_message += token
            else:
                # Not an insert.  Append the string to the message
                formatted_message += token
        return formatted_message
