# -*- coding: utf-8 -*-
"""
Implementation of 'managed_service_base' and related classes
"""
from __future__ import absolute_import

# Import python libs
import codecs
import datetime
import logging
import os
import socket
import sys
import threading

# Import local libs
# pylint: disable=wrong-import-position,import-error,no-name-in-module
import niskyline.messagebus.exceptions as exceptions
from niskyline.messagebus.skyline_environment_constants import NODE_NAME
from niskyline.messagebus.skyline_environment_constants import SERVICE_GUID
from niskyline.messagebus.skyline_environment_constants import SERVICE_GROUP_NAME
from niskyline.messagebus.skyline_environment_constants import DEFAULT_SERVICE_GROUP_NAME
import niskyline.messagebus.paths as paths
from niskyline.messagebus.error import Error
from niskyline.messagebus.process_logger import ProcessLogger
from niskyline.messagebus.message_service import MessageService
from niskyline.messagebus.message_service_builder import MessageServiceBuilder
from niskyline.messagebus.managed_service_builder import ManagedServiceBuilder
from niskyline.messagebus.message_subscriber import MessageSubscriber
from niskyline.messagebus.message_subscriber_builder import MessageSubscriberBuilder
import niskyline.messagebus.service_manager_messages as service_manager_messages
import niskyline.messagebus.configuration_messages as configuration_messages
import niskyline.messagebus.audit_trail_messages as audit_trail_messages
# pylint: enable=wrong-import-position,import-error,no-name-in-module

# Set up logging
LOGGER = logging.getLogger(__name__)


class ManagedServiceBase(object):  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """
    Managed Service Base.

    This class is meant to be inherited by a user created object. It
    enables integration with the Skyline Service Manager.
    """
    def __init__(self, service_name, shutdown_event, managed_service_builder=None):  # pylint: disable=too-many-branches,too-many-statements
        """
        :param service_name: The name of the message service. Ignored
            if ``managed_service_builder`` is not ``None``. In this
             case, ``service_name`` may be ``None``.
        :type service_name: str or None
        :param shutdown_event: A :class`threading.Semaphore` instance
            that is released when the Skyline Service Manager requires
            the service to shut down.
        :type shutdown_event: threading.Semaphore
        :param message_service_builder: A
            :class:`niskyline.messagebus.managed_service_builder.ManagedServiceBuilder`
            object used in the construction of this object. May be
            ``None``, in which case ``service_name`` must be set.
        :type message_service_builder:
            niskyline.messagebus.managed_service_builder.ManagedServiceBuilder
            or None
        """
        LOGGER.debug('ManagedServiceBase constructor!')

        if managed_service_builder is None:
            managed_service_builder = ManagedServiceBuilder(service_name)

        if shutdown_event is None:
            raise exceptions.SkylineException('Shutdown event cannot be None')

        self._closing = False
        self._work_subscriber = None
        self._control_subscriber = None
        self._callback_with_state = None
        self._node_name = ''
        self._service_name = None
        self._instance_name = None
        self._shutdown_thread = None
        self._old_stdout = None
        self._old_stderr = None
        self._devnull_fp = None

        if (sys.stdout is None or sys.stderr is None or
                sys.stdout.encoding != 'utf-8' or
                sys.stderr.encoding != 'utf-8'):
            # We will hit this case when the stdout/stderr pipes
            # are closed before creating this process. This is the
            # current behavior of Service Manager. In Python 3.5.x,
            # this will result in `sys.stdout` and `sys.stderr` being
            # None. In Python 3.6.x, this will result in `sys.stdout`
            # and `sys.stderr` being created by Python, but imperfectly
            # such that the encoding is the system encoding (typically
            # 'cp1252') instead of 'utf-8' which will still cause
            # problems. This should be fixed in Python 3.7.x in which
            # case it should create pipes with an encoding of 'utf-8'.
            # Once Python 3.7.x is the minimum supported version, this
            # code may be removed.
            self._old_stdout = sys.stdout
            self._old_stderr = sys.stderr
            self._devnull_fp = open(os.devnull, 'w', encoding='utf-8')
            sys.stdout = self._devnull_fp
            sys.stderr = self._devnull_fp

        self.set_service_name(managed_service_builder.get_service_name())
        self._status = service_manager_messages.ServiceState(
            service_manager_messages.ServiceState.STARTING
        )
        self._standalone = managed_service_builder.get_standalone_property()

        if not managed_service_builder.get_instance_name():
            self.set_instance_name(self.get_unique_instance_name())
        else:
            self.set_instance_name(managed_service_builder.get_instance_name())

        self._no_configuration_request = managed_service_builder.get_no_configuration_request()

        self._service_guid = os.environ.get(SERVICE_GUID)
        if not self._service_guid:
            self._service_guid = ''

        self._service_group_name = os.environ.get(SERVICE_GROUP_NAME)
        if not self._service_group_name:
            self._service_group_name = DEFAULT_SERVICE_GROUP_NAME

        self._receive_control_message_callback = ReceiveControlMessageCallback(self)  # pylint: disable=invalid-name

        if not self._standalone:
            message_subscriber_builder = MessageSubscriberBuilder(self.get_control_queue_name())
            message_subscriber_builder.set_callback(self._receive_control_message_callback)
            message_subscriber_builder.set_register_default_binding(True)
            message_subscriber_builder.set_auto_start_consumers(False)
            self._control_subscriber = MessageSubscriber(message_subscriber_builder)
            self._control_subscriber.register_broadcast_callback(
                service_manager_messages.SvcMgrSendServiceStatusRequestBroadcast.__name__,
                self._receive_control_message_callback)
            self._control_subscriber.register_message_name_callback(
                configuration_messages.ConfigurationGetSectionKeyValuesResponse.__name__,
                self._receive_control_message_callback)

        self._receive_response_message_callback = ReceiveResponseMessageCallback(self)  # pylint: disable=invalid-name
        self._receive_response_message_callback_with_state = (  # pylint: disable=invalid-name
            ReceiveResponseMessageCallbackWithState(self)
        )
        if not managed_service_builder.get_no_work_subscriber():
            if managed_service_builder.get_work_subscriber_builder():
                work_builder = managed_service_builder.get_work_subscriber_builder()
            else:
                work_builder = MessageSubscriberBuilder(self.get_service_name())

            work_builder.set_callback(self._receive_response_message_callback)
            work_builder.set_auto_start_consumers(False)
            self._work_subscriber = MessageSubscriber(work_builder)

        if managed_service_builder.get_instance_subscriber_builder() is not None:
            instance_builder = managed_service_builder.get_instance_subscriber_builder()
        else:
            instance_builder = MessageSubscriberBuilder(self.get_full_name())

        instance_builder.set_register_default_binding(True)
        instance_builder.set_auto_start_consumers(False)

        message_service_builder = MessageServiceBuilder(self.get_service_name())
        message_service_builder.set_instance_name(self.get_instance_name())
        message_service_builder.set_subscriber_builder(instance_builder)

        self._message_service = MessageService(message_service_builder=message_service_builder)
        self._message_service.set_instance_name(self.get_instance_name())

        self._process_logger = ProcessLogger(
            self.get_full_name(),
            None,
            self._message_service.get_publisher().publish_message_callback,
            self._message_service._register_broadcast_callback)  # pylint: disable=protected-access

        self._message_service.set_trace_logger(
            self._process_logger.make_trace_logger(self.get_service_name())
        )

        if self._control_subscriber is not None:
            self._control_subscriber.set_trace_logger(
                self._process_logger.make_trace_logger(
                    'Control',
                    self._message_service.get_trace_logger()))

        if self._work_subscriber is not None:
            self._work_subscriber.set_trace_logger(self._message_service.get_trace_logger())

        self._message_service.set_callback(self._receive_response_message_callback)
        self._message_service.set_callback_with_state(
            self._receive_response_message_callback_with_state
        )
        self._trace_unhandled_message = self.get_trace_logger().make_trace_point('UnhandledMessage')

        self._configured = False
        self._need_to_go_live = False
        self._trace_unhandled_message = None
        self._shutdown_event = shutdown_event
        self._shutdown_thread = None

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        """
        Close this instance of :class:`ManagedServiceBase`.
        """
        if self._closing:
            return
        self._closing = True
        self._wait_on_shutdown_thread()
        if self._work_subscriber is not None:
            self._work_subscriber.close()
        if self._control_subscriber is not None:
            self._control_subscriber.close()
        if self._message_service is not None:
            self._message_service.close()
        if self._devnull_fp is not None:
            sys.stdout = self._old_stdout
            sys.stderr = self._old_stderr
            self._devnull_fp.close()
            self._devnull_fp = None
            self._old_stdout = None
            self._old_stderr = None

    def get_service_guid(self):
        """
        Get the GUID (globally unique identifier) used for this service.

        :return: The GUID (globally unique identifier) used for this
            service.
        :rtype: str
        """
        return self._service_guid

    def set_service_guid(self, service_guid):
        """
        Set the GUID (globally unique identifier) used for this service.

        :param service_guid: The GUID (globally unique identifier) used
            for this service.
        :type service_guid: str
        """
        self._service_guid = service_guid

    def get_service_group_name(self):
        """
        Get the service group name.

        :return: the service group name.
        :rtype: str
        """
        return self._service_group_name

    def set_service_group_name(self, service_group_name):
        """
        Set the service group name.

        :param service_group_name: the service group name.
        :type service_group_name: str
        """
        self._service_group_name = service_group_name

    @staticmethod
    def get_unique_instance_name():
        """
        Generate a unique instance name.

        :return: A unique instance name.
        :rtype: str
        """
        return MessageService.get_unique_instance_name()

    def get_full_name(self, service_name=None, instance_name=None):
        """
        Get the full name based on service name and instance name if
        it exists.

        :param service_name: The name of the message service. If
            ``None``, will use the service name from this instance
            of class:`ManagedServiceBase`.
        :type service_name: str or None
        :param instance_name: The name of the message service instance.
            If ``None``, will use the instance name from this instance
            of class:`ManagedServiceBase`. Both may be ``None`` if
            there is only one instance of the service.
        :type instance_name: str or None
        :return: The full name.
        :rtype: str
        """
        if service_name is None:
            service_name = self._service_name
        if instance_name is None:
            instance_name = self._instance_name
        return MessageService.get_full_name(service_name=service_name, instance_name=instance_name)

    @staticmethod
    def get_host_name():
        """
        Get the hostname of this machine.

        :return: The hostname of this machine.
        :rtype: str
        """
        return socket.gethostname()

    def get_node_name(self):
        """
        Get the node name of the Service Manager instance that this
        service will use.

        :return: The node name of the Service Manager instance that this
            service will use.
        :rtype: str
        """
        if not self._node_name:
            node_name = ''
            try:
                node_name = os.environ[NODE_NAME]
            except KeyError:
                pass
            if node_name != '':
                self._node_name = node_name.lower()
            else:
                name = self.get_host_name().lower()
                config_dir = paths.get_configuration_directory()
                node_name_config_path = os.path.join('/', config_dir, 'NodeName.txt')
                try:
                    with codecs.open(node_name_config_path, 'r', encoding='utf-8-sig') as fp_:
                        file_contents = fp_.read().lower()
                        if file_contents is not None and file_contents != 'localhost':
                            name = file_contents
                except IOError:
                    pass
                self._node_name = name
        return self._node_name

    def set_node_name(self, node_name):
        """
        Set the node name of the Service Manager instance that this
        service will use.

        :param node_name: The node name of the Service Manager instance
             that this service will use.
        :type node_name: str
        """
        self._node_name = node_name

    def get_service_name(self):
        """
        Get the service name.

        :return: The service name.
        :rtype: str
        """
        return self._service_name

    def set_service_name(self, service_name):
        """
        Set the service name.

        :param service_name: The service name.
        :type service_name: str
        """
        self._service_name = service_name

    def get_instance_name(self):
        """
        Get the instance name.

        :return: The instance name. May be ``None``.
        :rtype: str or None
        """
        return self._instance_name

    def set_instance_name(self, instance_name):
        """
        Set the instance name.

        :param instance_name: The instance name. May be ``None``.
        :type instance_name: str or None
        """
        self._instance_name = instance_name

    def get_trace_logger(self):
        """
        Get the Trace Logger.

        :return: The Trace Logger. May be ``None``.
        :rtype: niskyline.messagebus.trace_logger.TraceLogger or None
        """
        return self._message_service.get_trace_logger()

    def make_trace_logger(self, name):
        """
        Create an instance of
        :class:`niskyline.messagebus.trace_logger.TraceLogger`.

        :param name: The last part of the name to use for this Trace Logger.
            Will not be the full name if there is an instance of
            :class:`niskyline.messagebus.trace_logger.TraceLogger`
            associated with this instance of :class:`ManagedServiceBase`.
            In that case, the name of that Trace Logger will be used as
            the parent name (the second last part of the name).
        :type name: str
        :return: A new instance of
            :class:`niskyline.messagebus.trace_logger.TraceLogger`.
        :rtype: niskyline.messagebus.trace_logger.TraceLogger
        """
        return self._process_logger.make_trace_logger(name, self.get_trace_logger())

    def set_callback_with_state(self, callback_with_state):
        """
        Set the callback object which supports user specific data (also
        known as ``state``) that is invoked whenever a message is
        received that was sent from this service or whenever a broadcast
        is received unless there is a more specific callback registered.

        :param callback_with_state: A class instance that contains a method
            named ``callback``. The ``callback`` function takes an argument of
            type :class:`niskyline.messagebus.generic_message.GenericMessage`
            and a second argument that could be any type for user specific
            data.
        :rtype: object
        """
        self._callback_with_state = callback_with_state

    def get_current_state(self):
        """
        Get the current state of the service.

        :return: The current state of the service.
        :rtype: niskyline.messagebus.service_manager_messages.ServiceState
        """
        return self._status

    def set_current_state(self, service_state):
        """
        Set the current state of the service.

        :param service_state: The current state of the service.
        :type service_state:
            niskyline.messagebus.service_manager_messages.ServiceState
        """
        if self._status.get_value() == service_state.get_value():
            return
        if self._status.get_value() == service_manager_messages.ServiceState.ERROR \
                and service_state.get_value() != service_manager_messages.ServiceState.STOPPED:
            return

        self._status = service_state

        if self._status.get_value() == service_manager_messages.ServiceState.REGISTERED:
            return
        service_status_broadcast = service_manager_messages.ServiceStatusBroadcast(
            self.get_node_name(),
            self.get_service_guid(),
            self.get_service_name(),
            self.get_service_group_name(),
            self.get_control_queue_name(),
            self.get_current_state())
        self.publish_broadcast(service_status_broadcast)

    def get_publisher(self):
        """
        Get the message publisher.

        :return: The message publisher.
        :rtype: niskyline.messagebus.message_publisher.MessagePublisher
        """
        return self._message_service.get_publisher()

    def get_work_subscriber(self):
        """
        Get the message subscriber for work-related messages.

        :return: The message subscriber for work-related messages.
        :rtype: niskyline.messagebus.message_subscriber.MessageSubscriber
        """
        return self._work_subscriber

    def get_instance_subscriber(self):
        """
        Get the message subscriber for messages for this specific
        service instance.

        :return: The message subscriber for messages for this specific
            service instance.
        :rtype: niskyline.messagebus.message_subscriber.MessageSubscriber
        """
        return self._message_service.get_subscriber()

    def get_message_service(self):
        """
        Get the associated message service.

        :return: The associated message service.
        :rtype: niskyline.messagebus.message_service.MessageService
        """
        return self._message_service

    def get_control_queue_name(self):
        """
        Get the name of the queue used to process control messages
        (messages coming from the Service Manager).

        :return: The name of the queue used to process control messages.
        :rtype: str
        """
        return self.get_full_name(self._service_name, self._instance_name) + '_Control'

    def _register_with_service_manager(self):
        """
        Perform registration with the Service Manager.
        """
        process_id = os.getpid()
        routed_message = \
            service_manager_messages.SvcMgrRegisterLocalServiceRoutedMessage(
                send_to='ServiceManager_' + self.get_node_name(),
                service_guid_=self.get_service_guid(),
                node_name_=self.get_node_name(),
                service_name_=self.get_service_name(),
                service_group_name_=self.get_service_group_name(),
                reply_to_queue_=self.get_control_queue_name(),
                process_id_=process_id
            )
        self.publish_routed_message(routed_message)
        self.set_current_state(
            service_manager_messages.ServiceState(service_manager_messages.ServiceState.REGISTERED)
        )

    def _send_service_status_broadcast(self):
        """
        Send the broadcast message that contains service status. It is
        an instance of
        :class:`niskyline.messagebus.service_manager_messages.ServiceStatusBroadcast`.
        """
        service_status_broadcast = service_manager_messages.ServiceStatusBroadcast(
            self.get_node_name(),
            self.get_service_guid(),
            self.get_service_name(),
            self.get_service_group_name(),
            self.get_control_queue_name(),
            self.get_current_state()
        )
        self.publish_routed_message(service_status_broadcast)

    def publish_audit_entry(self, subtype):
        """
        Publish an audit message. It is an instance of
        :class:`niskyline.messagebus.audit_trail_messages.AuditEntry`.

        :param subtype: The audit entry subtype.
        :type subtype: str
        """
        audits = []
        audit_entry = audit_trail_messages.AuditEntry(
            timestamp_=datetime.datetime.utcnow(),
            subtype_=subtype,
            originator_=self.get_service_name(),
            user_name_=None,
            record_type_='Service',
            device_=None,
            detail_=None,
            int1_=0,
            int2_=0,
            int3_=0,
            int4_=0,
            str1_=None,
            str2_=None,
            str3_=None,
            str4_=None
        )
        audits.append(audit_entry)
        audit_add_routed_message = audit_trail_messages.AuditAddRoutedMessage(audits)
        self.publish_routed_message(audit_add_routed_message)

    def _receive_message(self, generic_message):
        """
        The default handler for when a message is received on the work
        or instance queue. Logs an error because it is expected that a
        more specific handler will be registered to handle the message.

        :param generic_message: The incoming message to handle. The
            class of the message is
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type generic_message: niskyline.messagebus.generic_message.GenericMessage
        """
        if generic_message.has_error():
            log_message = 'Unhandled message ' + \
                          generic_message.get_message_name() + \
                          ': Origin = ' + generic_message.get_origin() + \
                          ' Error = ' + generic_message.get_error
            self.get_trace_logger().log_error(log_message)
        else:
            log_message = generic_message.get_message_name() + \
                          ": Origin=" + generic_message.get_origin()
            self.get_trace_logger().log_error(log_message)

    def _try_set_configuration(self, sections):  # pylint: disable=no-self-use,unused-argument
        """
        Not implemented. Always returns success.

        :param sections: A list of key/value pair dictionaries for each
            section in the configuration.
        :type sections: list(niskyline.messagebus.configuration_messages.SectionKeyValues)
        :return: An instance of
            :class:`niskyline.messagebus.error.Error` representing
            the return status of this function.
        :rtype: niskyline.messagebus.error.Error
        """
        # TODO: Implement this function when a service needs to perform
        #       initialization using the Configuration service.
        # NOTE: This function is called before Message subscriptions
        #       are enabled.
        error = Error()
        return error

    def initialize(self):
        """
        Initialize this instance of :class:`ManagedServiceBase`. Should
        be invoked by the ``__init__`` method of the parent class.

        Supports both standalone and integrated with Service Manager
        use cases.
        """
        if self._standalone:
            self.go_live()
        else:
            self._control_subscriber.start_handling_messages()
            if not self._no_configuration_request:
                request = \
                    configuration_messages.ConfigurationGetSectionKeyValuesRequest()
                request.get_header().set_reply_to(self.get_control_queue_name())
                self.get_publisher().publish_message_base(request)
            else:
                self._configured = True
            self._register_with_service_manager()

    def publish_request(self, request_message, state=None):
        """
        Publish a request message.

        :param request_message: The request message.
        :type request_message:
            niskyline.messagebus.request_message.RequestMessage
        :param state: Optional user specific data. May be any
            type. Used if a callback with state is registered.
        :type state: object or None
        """
        self._message_service.publish_request(request_message, state)

    def publish_response(self, response_message):
        """
        Publish a response message.

        :param response_message: The response message.
        :type response_message:
            niskyline.messagebus.response_message.ResponseMessage
        """
        self._message_service.publish_response(response_message)

    def publish_routed_message(self, routed_message, ignore_response=False):
        """
        Publish a routed message.

        :param routed_message: The routed message.
        :type routed_message:
            niskyline.messagebus.routed_message.RoutedMessage
        :param ignore_response: The ``IgnoreResponse`` property in the
            message header.
        :type ignore_response: bool
        """
        self._message_service.publish_routed_message(routed_message, ignore_response)

    def publish_broadcast(self, broadcast_message):
        """
        Publish a broadcast message.

        :param broadcast_message: The broadcast message.
        :type broadcast_message:
            niskyline.messagebus.broadcast_message.BroadcastMessage
        """
        self._message_service.publish_broadcast(broadcast_message)

    def publish_generic_message(self, generic_message, state=None):
        """
        Publish a generic message.

        :param generic_message: The generic message.
        :type generic_message:
            niskyline.messagebus.generic_message.GenericMessage
        """
        if state:
            self._message_service.publish_request(generic_message, state)
        else:
            self._message_service.publish_generic_message(generic_message)

    def publish_synchronous_message(self, message, mandatory=False, timeout_seconds=None):
        """
        Synchronously publish a message.

        :param message: The message object. The class of this object
            inherits from
            :class:`niskyline.messagebus.message_base.MessageBase`.
        :type message: niskyline.messagebus.message_base.MessageBase
        :param mandatory: The mandatory delivery notification flag.
        :type mandatory: bool
        :param timeout_seconds: The timeout in seconds for this query.
            If ``None``, will use the timeout specified by the
            :class:`niskyline.messagebus.message_service_builder.MessageServiceBuilder`
            object used to create this instance of
            :class:`MessageService`.
        :type timeout_seconds: int or float
        :return: The response message.
        :rtype: niskyline.messagebus.generic_message.GenericMessage
        """
        return self._message_service.publish_synchronous_message(
            message, mandatory=mandatory, timeout_seconds=timeout_seconds
        )

    def publish_generic_response(self, routed_message, error=None):
        """
        Publish a generic response.

        :param request: The request message that is being responded to.
        :type request:
            niskyline.messagebus.routed_message.RoutedMessage
        """
        self._message_service.publish_generic_response(routed_message, error=error)

    def register_work_subscriber_callback(self, message_name, callback):  # pylint: disable=invalid-name
        """
        Register a callback that will be invoked when a
        certain message name is used in a response message.
        Used for work queue messages.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._work_subscriber.register_message_name_callback(
            message_name, callback, False
        )

    def register_instance_subscriber_callback(self, message_name, callback):  # pylint: disable=invalid-name
        """
        Register a callback that will be invoked when a
        certain message name is used in a response message.
        Used for instance queue messages.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._message_service.get_subscriber().register_message_name_callback(
            message_name, callback, False
        )

    def register_response_callback(self, message_name, response_callback):
        """
        Register a callback that will be invoked when a
        certain message name is used in a response message.
        Used for instance queue messages.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._message_service.register_response_callback(message_name, response_callback)

    def register_broadcast_callback(self, message_name, callback):
        """
        Register a callback that will be invoked when a
        certain message name is used in a broadcast message.
        Used for work queue messages.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._work_subscriber.register_broadcast_callback(message_name, callback)

    def register_broadcast_for_each_instance(self, message_name, callback):  # pylint: disable=invalid-name
        """
        Register a callback that will be invoked when a
        certain message name is used in a broadcast message.
        Used for instance queue messages.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._message_service.register_broadcast_callback(message_name, callback)

    def go_live(self):
        """
        Notify the Service Manager that this service is live.
        """
        LOGGER.debug('Managed Service Base: go live!')
        process_id = os.getpid()

        log_message = 'Go Live ' + str(process_id) + ' - ' \
                      + self.get_instance_name() \
                      + self.get_service_name() \
                      + ' ServiceGuid='

        if self._standalone:
            log_message += '(Standalone)'
        if self.get_service_guid():
            log_message += self.get_service_guid()

        self.get_trace_logger().log_info(log_message)
        self.publish_audit_entry('Live')
        self._message_service.get_subscriber().start_handling_messages()
        if self._work_subscriber:
            self._work_subscriber.start_handling_messages()
        self.set_current_state(
            service_manager_messages.ServiceState(service_manager_messages.ServiceState.LIVE)
        )

    def make_idle(self):
        """
        Notify the Service Manager that this service is idle.
        """
        LOGGER.debug('Managed Service Base: make idle!')
        process_id = os.getpid()

        log_message = (
            'Idled ' + str(process_id) + ' - ' +
            self.get_instance_name() +
            self.get_service_name() +
            ' ServiceGuid='
        )

        if self._standalone:
            log_message += '(Standalone)'
        if self.get_service_guid():
            log_message += self.get_service_guid()

        self.get_trace_logger().log_info(log_message)
        self.publish_audit_entry('Idle')
        self._message_service.get_subscriber().stop_handling_messages()
        if self._work_subscriber:
            self._work_subscriber.stop_handling_messages()
        self.set_current_state(
            service_manager_messages.ServiceState(service_manager_messages.ServiceState.IDLE)
        )

    def _shutdown_thread_func(self):
        """
        The thread function for the shutdown thread.

        It will signal the shutdown event and tell the Service Manager
        that it's state is stopped.
        """
        self._shutdown_event.release()
        self.set_current_state(
            service_manager_messages.ServiceState(service_manager_messages.ServiceState.STOPPED)
        )

    def _start_shutdown_thread(self):
        """
        Start the shutdown thread if it wasn't already started.
        """
        if self._shutdown_thread is not None:
            return
        self._shutdown_thread = threading.Thread(
            target=self._shutdown_thread_func
        )
        self._shutdown_thread.daemon = True
        self._shutdown_thread.start()

    def _wait_on_shutdown_thread(self):
        """
        Wait for the shutdown thread to complete if it was started.
        """
        if self._shutdown_thread is None:
            return
        self._shutdown_thread.join()
        self._shutdown_thread = None


class ReceiveControlMessageCallback(object):  # pylint: disable=too-few-public-methods
    """
    Used to handle incoming control messages from the Service Manager.
    """
    def __init__(self, managed_service_base):
        """
        :param managed_service_base: The associated managed service.
        :type managed_service_base: ManagedServiceBase
        """
        self._managed_service_base = managed_service_base

    def callback(self, generic_message):  # pylint: disable=too-many-branches
        """
        The callback function invoked by the framework once a
        control message is received.

        :param generic_message: The response message.
        :type generic_message:
            niskyline.messagebus.generic_message.GenericMessage
        """
        message_name = generic_message.get_message_name()
        LOGGER.debug('Received message: %s', message_name)

        if (message_name ==
                service_manager_messages.SvcMgrSendServiceStatusRequestBroadcast.__name__):
            self._managed_service_base._send_service_status_broadcast()  # pylint: disable=protected-access
        elif (message_name ==
              configuration_messages.ConfigurationGetSectionKeyValuesResponse.__name__):
            LOGGER.debug('ConfigurationGetSectionKeyValuesResponse!')
            response = (
                configuration_messages.ConfigurationGetSectionKeyValuesResponse.from_message(
                    generic_message
                )
            )
            error = self._managed_service_base._try_set_configuration(  # pylint: disable=protected-access
                response.key_values)
            if error.has_value():
                LOGGER.debug('Has Error!')
                self._managed_service_base.get_trace_logger().log_error(error)
                self._managed_service_base.set_current_state(
                    service_manager_messages.ServiceState(
                        service_manager_messages.ServiceState.ERROR
                    )
                )
            else:
                LOGGER.debug('No Error.')
                self._managed_service_base._configured = True  # pylint: disable=protected-access
                if self._managed_service_base._need_to_go_live:  # pylint: disable=protected-access
                    self._managed_service_base.go_live()
        elif message_name == service_manager_messages.ServiceGoLiveRoutedMessage.__name__:
            LOGGER.debug('ServiceGoLiveRoutedMessage')
            service_go_live_routed_message = (
                service_manager_messages.ServiceGoLiveRoutedMessage.from_message(
                    send_to='',
                    message=generic_message
                )
            )
            if not self._managed_service_base.get_service_guid():
                self._managed_service_base.set_service_guid(
                    service_go_live_routed_message.service_guid
                )
            elif (self._managed_service_base.get_service_guid() !=
                  service_go_live_routed_message.service_guid):
                log_message = 'Go live request with mis-matched GUID: '
                log_message += self._managed_service_base.get_service_guid()
                log_message += " "
                log_message += service_go_live_routed_message.service_guid

                self._managed_service_base.get_trace_logger.log_error(log_message)
            if self._managed_service_base.get_current_state().get_value() \
                    != service_manager_messages.ServiceState.ERROR:
                if self._managed_service_base._configured:  # pylint: disable=protected-access
                    self._managed_service_base.go_live()
                else:
                    self._managed_service_base._need_to_go_live = True  # pylint: disable=protected-access
        elif message_name == service_manager_messages.ServiceMakeIdleRoutedMessage.__name__:
            if (self._managed_service_base.get_current_state() !=
                    service_manager_messages.ServiceState.ERROR):
                self._managed_service_base._need_to_go_live = False  # pylint: disable=protected-access
                self._managed_service_base.make_idle()
        elif message_name == service_manager_messages.ServiceShutdownRoutedMessage.__name__:
            self._managed_service_base._start_shutdown_thread()  # pylint: disable=protected-access
        else:
            log_message = 'Unhandled Control Message | '
            log_message += generic_message.get_message_name()
            self._managed_service_base.get_trace_logger().log_error(log_message)


class ReceiveResponseMessageCallback(object):  # pylint: disable=too-few-public-methods
    """
    Used to handle response messages without ``state`` (user specific
    data) on the work or instance queue.

    Logs an error because it is expected that a more specific handler
    will be registered to handle the message.
    """
    def __init__(self, managed_service_base):
        """
        :param managed_service_base: The associated managed service.
        :type managed_service_base: ManagedServiceBase
        """
        self._managed_service_base = managed_service_base

    def callback(self, generic_message):
        """
        The callback function invoked by the framework once a
        response message is received.

        :param generic_message: The response message.
        :type generic_message:
            niskyline.messagebus.generic_message.GenericMessage
        """
        self._managed_service_base._receive_message(generic_message)  # pylint: disable=protected-access


class ReceiveResponseMessageCallbackWithState(object):  # pylint: disable=too-few-public-methods
    """
    Used to handle response messages with ``state`` (user specific
    data) on the work or instance queue.

    Logs an error because it is expected that a more specific handler
    will be registered to handle the message.
    """
    def __init__(self, managed_service_base):
        """
        :param managed_service_base: The associated managed service.
        :type managed_service_base: ManagedServiceBase
        """
        self._managed_service_base = managed_service_base

    def callback(self, generic_message, state):  # pylint: disable=unused-argument
        """
        The callback function invoked by the framework once a
        response message is received.

        :param generic_message: The response message.
        :type generic_message:
            niskyline.messagebus.generic_message.GenericMessage
        """
        self._managed_service_base._receive_message(generic_message)  # pylint: disable=protected-access
