# -*- coding: utf-8 -*-
"""
Implementation of 'MessageService' and related classes
"""
from __future__ import absolute_import

# Import python libs
import logging
import sys
import uuid
# pylint: disable=import-error
if sys.version_info[0] < 3:
    from Queue import Queue
else:
    from queue import Queue
# pylint: enable=import-error

# Import local libs
# pylint: disable=wrong-import-position,import-error
import niskyline.messagebus.exceptions as exceptions
from niskyline.messagebus.generic_message import GenericMessage
from niskyline.messagebus.generic_response import GenericResponse
from niskyline.messagebus.message_publisher import MessagePublisher
from niskyline.messagebus.message_publisher_builder import MessagePublisherBuilder
from niskyline.messagebus.message_service_builder import MessageServiceBuilder
from niskyline.messagebus.message_subscriber import MessageSubscriber, SynchCallbackInfo
from niskyline.messagebus.message_subscriber_builder import MessageSubscriberBuilder
from niskyline.messagebus.routed_message import RoutedMessage
# pylint: enable=wrong-import-position,import-error

# Set up logging
LOGGER = logging.getLogger(__name__)


class MessageService(object):  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """
    This class oversees the overall usage of the message bus and is
    the top-level class used by Skyline-based clients and services.
    """
    def __init__(self, message_service_builder):
        """
        :param message_service_builder: A
            :class:`niskyline.messagebus.message_service_builder.MessageServiceBuilder`
            object used in the construction of this object.
        :type message_service_builder:
            niskyline.messagebus.message_service_builder.MessageServiceBuilder
        """
        LOGGER.debug('MessageService constructor!')
        self._closing = False
        self._subscriber = None
        self._trace_logger = None
        self._callback = None
        self._callback_with_state = None
        self._internal_response_callback = MessageServiceResponseCallback(self)
        self._response_callbacks = {}
        self._register_broadcast_callback = MessageServiceRegisterBroadcastCallback(self)
        self._service_name = message_service_builder.get_service_name()
        LOGGER.debug('service_name = %s', self._service_name)
        if not message_service_builder.get_instance_name():
            self._instance_name = self.get_unique_instance_name()
        else:
            self._instance_name = message_service_builder.get_instance_name()

        LOGGER.debug('_instance_name = %s', self._instance_name)

        self._request_timeout = message_service_builder.get_request_timeout()
        message_publisher_builder = MessagePublisherBuilder(self._service_name)
        message_publisher_builder.set_connection_manager(
            message_service_builder.get_connection_manager()
        )
        message_publisher_builder.set_reply_to(self.get_implicit_full_name())

        LOGGER.debug('fullName = %s', self.get_implicit_full_name())

        self._publisher = MessagePublisher(message_publisher_builder)

        subscriber_builder = message_service_builder.get_subscriber_builder()
        if subscriber_builder is None:
            subscriber_builder = MessageSubscriberBuilder(self.get_implicit_full_name())
            subscriber_builder.set_connection_manager(
                message_service_builder.get_connection_manager()
            )
            subscriber_builder.set_register_default_binding(True)

            if subscriber_builder.get_callback() is None:
                subscriber_builder.set_callback(MessageServiceBroadcastCallback(self))
            else:
                subscriber_builder.set_callback(subscriber_builder.get_callback())

        self._subscriber = MessageSubscriber(subscriber_builder)

    @classmethod
    def from_service_name(cls, service_name):
        """
        Create an instance of :class:`MessageService` based on the
        service name.

        :param service_name: The name of the message service.
        :type service_name: str
        :return: An instance of :class:`MessageService`.
        :rtype: MessageService
        """
        message_service_builder = MessageServiceBuilder(service_name)
        return cls(message_service_builder)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        """
        Close this instance of :class:`MessageService`.
        """
        if self._closing:
            return
        self._closing = True
        # `self._subscriber` and `self._publisher` may be used as part
        # of a `AmqpConnectionManager` singleton and hence are shared.
        # Due to this, don't close them explicitly here. Instead, rely
        # on them to close themselves once their reference count
        # reaches zero.
        self._subscriber = None
        self._publisher = None

    @staticmethod
    def get_unique_instance_name():
        """
        Generate a unique instance name.

        :return: A unique instance name.
        :rtype: str
        """
        return str(uuid.uuid4())[:8]

    def get_implicit_full_name(self):
        """
        Get the full name of this instance of :class:`MessageService`
        which takes into consideration the service name and the
        instance name if it is set.

        :return: The full name.
        :rtype: str
        """
        return self.get_full_name(self._service_name, self._instance_name)

    @staticmethod
    def get_full_name(service_name, instance_name=None):
        """
        Get the full name based on service name and instance name if
        it exists.

        :param service_name: The name of the message service.
        :type service_name: str
        :param instance_name: The name of the message service instance.
            May be ``None`` if there is only one instance of the
            message service.
        :type instance_name: str or None
        :return: The full name.
        :rtype: str
        """
        if service_name is None:
            raise exceptions.SkylineException('Service name cannot be none.')
        if instance_name is None:
            return service_name
        return service_name + '_' + instance_name

    def get_service_name(self):
        """
        Get the service name.

        :return: The service name.
        :rtype: str
        """
        return self._service_name

    def set_service_name(self, service_name):
        """
        Set the service name.

        :param service_name: The service name.
        :type service_name: str
        """
        self._service_name = service_name

    def get_instance_name(self):
        """
        Get the instance name.

        :return: The instance name. May be ``None``.
        :rtype: str or None
        """
        return self._instance_name

    def set_instance_name(self, instance_name):
        """
        Set the instance name.

        :param instance_name: The instance name. May be ``None``.
        :type instance_name: str or None
        """
        self._instance_name = instance_name

    def get_trace_logger(self):
        """
        Get the Trace Logger.

        :return: The Trace Logger.
        :rtype: niskyline.messagebus.trace_logger.TraceLogger
        """
        if self._trace_logger is None:
            raise exceptions.SkylineException('InvalidOperationException: No TraceLogger set.')
        return self._trace_logger

    def set_trace_logger(self, trace_logger):
        """
        Set the Trace Logger.

        :param trace_logger: The Trace Logger.
        :type trace_logger: niskyline.messagebus.trace_logger.TraceLogger
        """
        self._trace_logger = trace_logger
        self.get_subscriber().set_trace_logger(trace_logger)

    def get_request_timeout(self):
        """
        Get the request timeout (in seconds).

        :return: The request timeout (in seconds).
        :rtype: int or float
        """
        return self._request_timeout

    def set_request_timeout(self, request_timeout):
        """
        Set the request timeout (in seconds).

        :param request_timeout: The request timeout (in seconds).
        :type request_timeout: int or float
        """
        self._request_timeout = request_timeout

    def get_publisher(self):
        """
        Get the message publisher.

        :return: The message publisher.
        :rtype: niskyline.messagebus.message_publisher.MessagePublisher
        """
        return self._publisher

    def get_subscriber(self):
        """
        Get the message subscriber.

        :return: The message subscriber.
        :rtype: niskyline.messagebus.message_subscriber.MessageSubscriber
        """
        return self._subscriber

    def get_callback(self):
        """
        Get the callback object that is invoked whenever a message is
        received that was sent from this service or whenever a broadcast
        is received unless there is a more specific callback registered.

        :return: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :rtype: object
        """
        return self._callback

    def set_callback(self, callback):
        """
        Set the callback object that is invoked whenever a message is
        received that was sent from this service or whenever a broadcast
        is received unless there is a more specific callback registered.

        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._callback = callback

    def get_callback_with_state(self):
        """
        Get the callback object which supports user specific data (also
        known as ``state``) that is invoked whenever a message is
        received that was sent from this service or whenever a broadcast
        is received unless there is a more specific callback registered.

        :return: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage` and
            a second argument that could be any type for user specific data.
        :rtype: object
        """
        return self._callback_with_state

    def set_callback_with_state(self, callback_with_state):
        """
        Set the callback object which supports user specific data (also
        known as ``state``) that is invoked whenever a message is
        received that was sent from this service or whenever a broadcast
        is received unless there is a more specific callback registered.

        :param callback_with_state: A class instance that contains a method
            named ``callback``. The ``callback`` function takes an argument of
            type :class:`niskyline.messagebus.generic_message.GenericMessage`
            and a second argument that could be any type for user specific
            data.
        :rtype: object
        """
        self._callback_with_state = callback_with_state

    def publish_request(self, request_message, state=None):
        """
        Publish a request message.

        :param request_message: The request message.
        :type request_message:
            niskyline.messagebus.request_message.RequestMessage
        :param state: Optional user specific data. May be any
            type. Used if a callback with state is registered.
        :type state: object or None
        """
        if state is not None:
            self._subscriber.register_correlation_id_callback(
                request_message.get_correlation_id(),
                self._internal_response_callback,
                state
            )
        self._publisher.publish_message_base(request_message)

    def publish_response(self, response_message):
        """
        Publish a response message.

        :param response_message: The response message.
        :type response_message:
            niskyline.messagebus.response_message.ResponseMessage
        """
        self._publisher.publish_message_base(response_message)

    def publish_routed_message(self, routed_message, ignore_response=True):
        """
        Publish a routed message.

        :param routed_message: The routed message.
        :type routed_message:
            niskyline.messagebus.routed_message.RoutedMessage
        :param ignore_response: The ``IgnoreResponse`` property in the
            message header.
        :type ignore_response: bool
        """
        routed_message.set_ignore_response(ignore_response)
        self._publisher.publish_message_base(routed_message)

    def publish_broadcast(self, broadcast_message):
        """
        Publish a broadcast message.

        :param broadcast_message: The broadcast message.
        :type broadcast_message:
            niskyline.messagebus.broadcast_message.BroadcastMessage
        """
        self._publisher.publish_message_base(broadcast_message)

    def publish_generic_message(self, generic_message):
        """
        Publish a generic message.

        :param generic_message: The generic message.
        :type generic_message:
            niskyline.messagebus.generic_message.GenericMessage
        """
        self._publisher.publish_generic_message(generic_message)

    def publish_generic_request(self, generic_message, state=None):
        """
        Publish a generic message as a request.

        :param generic_message: The generic message.
        :type generic_message:
            niskyline.messagebus.generic_message.GenericMessage
        :param state: Optional user specific data. May be any
            type. Used if a callback with state is registered.
        :type state: object or None
        """
        self._subscriber.register_correlation_id_callback(
            generic_message.get_correlation_id(),
            self._internal_response_callback,
            state
        )
        self._publisher.publish_generic_message(generic_message)

    def publish_synchronous_message(self, message, mandatory=False, timeout_seconds=None):
        """
        Synchronously publish a message.

        :param message: The message object. The class of this object
            inherits from
            :class:`niskyline.messagebus.message_base.MessageBase`.
        :type message: niskyline.messagebus.message_base.MessageBase
        :param mandatory: The mandatory delivery notification flag.
        :type mandatory: bool
        :param timeout_seconds: The timeout in seconds for this query.
            If ``None``, will use the timeout specified by the
            :class:`niskyline.messagebus.message_service_builder.MessageServiceBuilder`
            object used to create this instance of
            :class:`MessageService`.
        :type timeout_seconds: int or float
        :return: The response message.
        :rtype: niskyline.messagebus.generic_message.GenericMessage
        """
        if timeout_seconds is None:
            timeout_seconds = self._request_timeout
        if issubclass(message.__class__, RoutedMessage):
            message.set_ignore_response(False)
        return self._publish_synchronous_request(message, timeout_seconds, mandatory)

    def publish_generic_response(self, request, error=None):
        """
        Publish a generic response.

        :param request: The request message that is being responded to.
        :type request:
            niskyline.messagebus.routed_message.RoutedMessage
        """
        self._publish_generic_response(request, error)

    def register_response_callback(self, message_name, callback):
        """
        Register a callback that will be invoked when a
        certain message name is used in a response message.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._response_callbacks[message_name] = callback

    def unregister_response_callback(self, message_name):
        """
        Unregister a callback that would have been invoked when a
        certain message name is used in a response message.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        """
        callback = self._response_callbacks.pop(message_name, None)
        if callback is None:
            if self._trace_logger is not None:
                self._trace_logger.log_error(
                    'Failed to remove register message name callback | ' +
                    message_name
                )

    def register_broadcast_callback(self, message_name, callback):
        """
        Register a callback that will be invoked when a
        certain message name is used in a broadcast message.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._register_broadcast_callback.register_broadcast_callback(message_name, callback)

    def unregister_broadcast_callback(self, message_name):
        """
        Unregister a callback that would have been invoked when a
        certain message name is used in a broadcast message.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        """
        self._subscriber.unregister_message_name_callback(message_name)
        params = []
        params.append(message_name)
        self._subscriber.unregister_binding(params)

    def _publish_synchronous_request(self, request, timeout_seconds, mandatory=False):
        """
        Synchronously publish a request.

        :param request: The message object. The class of this object
            inherits from
            :class:`niskyline.messagebus.message_base.MessageBase` or is
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type request: niskyline.messagebus.message_base.MessageBase
            or niskyline.messagebus.generic_message.GenericMessage
        :param timeout_seconds: The timeout in seconds for this query.
        :type timeout_seconds: int or float
        :param mandatory: The mandatory delivery notification flag.
        :type mandatory: bool
        :return: The response message.
        :rtype: niskyline.messagebus.generic_message.GenericMessage
        """
        queue = Queue()
        info = SynchCallbackInfo()
        info.set_queue(queue)

        self._subscriber._synch_callbacks[request.get_correlation_id()] = info  # pylint: disable=protected-access

        if issubclass(request.__class__, GenericMessage):
            self._publisher.publish_generic_message(request, mandatory)
        else:
            self._publisher.publish_message_base(request, mandatory)

        try:
            generic_message = self._subscriber.wait_for_message(queue, True, timeout_seconds)
        except exceptions.SkylineException:
            self._subscriber._synch_callbacks.pop(request.get_correlation_id(), None)  # pylint: disable=protected-access
            raise
        self._subscriber._synch_callbacks.pop(request.get_correlation_id(), None)  # pylint: disable=protected-access

        if self._trace_logger is not None and generic_message is not None:
            self._trace_logger.log_error(
                'Service did not respond | ' +
                request.get_routing_key()
            )
        return generic_message

    def _publish_generic_response(self, request, error):
        """
        Publish a generic response.

        :param request: The request message that is being responded to.
        :type request:
            niskyline.messagebus.routed_message.RoutedMessage
        """
        has_error = error.has_value()
        if not request.get_ignore_response():
            if not has_error:
                generic_response = GenericResponse.from_routing_source(request)
                self.publish_generic_response(generic_response)
            else:
                generic_response = GenericResponse.from_routing_source(request, error)
                self.publish_generic_response(generic_response)
        elif has_error:
            self._trace_logger.log_error(
                'Routed Message had error but response was ignored | ' +
                request.get_message_name() +
                ' from ' +
                request.get_origin() +
                ' | Error: ' +
                str(error)
            )


class MessageServiceRegisterBroadcastCallback(object):  # pylint: disable=too-few-public-methods
    """
    Used to give access to registering a broadcast callback.

    Used in :class:`MessageService` and
    :class:`niskyline.messagebus.process_logger.ProcessLogger`.
    """
    def __init__(self, message_service):
        """
        :param message_service: The associated message service.
        :type message_service: MessageService
        """
        self._message_service = message_service

    def register_broadcast_callback(self, message_name, callback):
        """
        Register a callback that will be invoked when a
        certain message name is used in a broadcast message.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        params = []
        params.append(message_name)
        self._message_service._subscriber.register_binding(params)  # pylint: disable=protected-access
        self._message_service._subscriber.register_message_name_callback(message_name, callback)  # pylint: disable=protected-access


class MessageServiceResponseCallback(object):  # pylint: disable=too-few-public-methods
    """
    Used to allow invoking user defined callbacks for response messages
    at various levels.

    Only one is invoked per message, in the following priority:
        1) Callbacks specific to the message name.
        2) General callbacks with ``state`` (user defined data).
        3) General callbacks without ``state`` (user defined data).
    """
    def __init__(self, message_service):
        """
        :param message_service: The associated message service.
        :type message_service: MessageService
        """
        self._message_service = message_service

    def callback(self, generic_message, state=None):
        """
        The callback function invoked by the framework once a
        response message is received.

        :param generic_message: The response message.
        :type generic_message:
            niskyline.messagebus.generic_message.GenericMessage
        """
        response_callback = self._message_service._response_callbacks.get(  # pylint: disable=protected-access
            generic_message.get_message_name()
        )
        if response_callback is not None:
            response_callback.callback(generic_message, state)
        elif self._message_service._callback_with_state is not None and state is not None:  # pylint: disable=protected-access
            self._message_service._callback_with_state.callback(generic_message, state)  # pylint: disable=protected-access
        elif self._message_service._callback is not None:  # pylint: disable=protected-access
            self._message_service._callback.callback(generic_message)  # pylint: disable=protected-access
        else:
            if self._message_service._trace_logger is not None:  # pylint: disable=protected-access
                self._message_service._trace_logger.log_error(  # pylint: disable=protected-access
                    'No callback registered for response | ' +
                    generic_message.get_message_name()
                )


class MessageServiceBroadcastCallback(object):  # pylint: disable=too-few-public-methods
    """
    Used to allow invoking user defined callbacks for broadcast messages
    at various levels.

    Only one is invoked per message, in the following priority:
        1) Callbacks specific to the message name.
        2) General callbacks without ``state`` (user defined data).
    """
    def __init__(self, message_service):
        """
        :param message_service: The associated message service.
        :type message_service: MessageService
        """
        self._message_service = message_service
        self._response_callback = MessageServiceResponseCallback(message_service)

    def callback(self, generic_message):
        """
        The callback function invoked by the framework once a
        broadcast message is received.

        :param generic_message: The broadcast message.
        :type generic_message:
            niskyline.messagebus.generic_message.GenericMessage
        """
        self._response_callback.callback(generic_message, None)
