# -*- coding: utf-8 -*-
"""
Implementation of 'MessageSubscriber' and related classes
"""
from __future__ import absolute_import

# Import python libs
import datetime
import logging
import threading
import sys
import time
# pylint: disable=import-error
if sys.version_info[0] < 3:
    from Queue import Queue, Empty
else:
    from queue import Queue, Empty
# pylint: enable=import-error

# Import local libs
# pylint: disable=wrong-import-position,import-error
from niskyline.messagebus.message_subscriber_builder import MessageSubscriberBuilder
from niskyline.messagebus.amqp_connection_manager import AmqpConnectionManager
from niskyline.messagebus.amqp_consumer_builder import AmqpConsumerBuilder
from niskyline.messagebus.amqp_consumer import AmqpConsumer
from niskyline.messagebus.trace_point import TracePoint
# pylint: enable=wrong-import-position,import-error

# Set up logging
LOGGER = logging.getLogger(__name__)

if sys.version_info[0] < 3:
    # Python 2 does not have built-in support for the monotonic clock.
    # We will use the system clock in this case, which is unsafe since the
    # clock can be changed mid-operation. Eventually, we will drop Python 2
    # support and this logic can disappear.
    USE_MONOTONIC_CLOCK = False
else:
    USE_MONOTONIC_CLOCK = True


class MessageSubscriber(object):  # pylint: disable=too-many-instance-attributes
    """
    The Message Publisher.

    This class is used to subscribe to and consume messages.
    """
    def __init__(self, builder=None):
        """
        :param builder: A
            :class:`niskyline.messagebus.message_subsciber_builder.MessageSubscriberBuilder`
            object used in the construction of this object. May be ``None`` if default
            behavior is desired.
        :type builder:
            niskyline.messagebus.message_subsciber_builder.MessageSubscriberBuilder
        """
        self._closing = False
        self._amqp_connection_manager = builder.get_connection_manager()
        self._amqp_consumers = []
        self._msg_name_callbacks = {}
        self._corr_id_callbacks = {}
        self._synch_callbacks = {}
        self._queue_name = builder.get_queue_name()
        self._explicit_ack = builder.get_explicit_ack()
        self._correlation_id_timeout = builder.get_correlation_id_timeout()
        self._local_message_queue = None

        if USE_MONOTONIC_CLOCK:
            clock_info = time.get_clock_info('monotonic')  # pylint: disable=no-member
            if clock_info.adjustable:
                self._clock_res = clock_info.resolution
            else:
                self._clock_res = 1
            self._last_correlation_id_check = time.monotonic()  # pylint: disable=no-member
        else:
            self._clock_res = None
            self._last_correlation_id_check = datetime.datetime.utcnow()

        self._callback = builder.get_callback()
        self._trace_unhandled_message = TracePoint('UnhandledMessage')
        self._trace_logger = builder.get_trace_logger()
        if self._trace_logger is not None:
            self._trace_unhandled_message = self._trace_logger.make_trace_point('UnhandledMessage')
        self._message_handling_threads = []
        if builder.get_local_queue_task_threads() > 1:
            self._local_message_queue = Queue(builder.get_max_local_queue_depth())
            for _ in range(builder.get_local_queue_task_threads()):
                thread = threading.Thread(target=self._process_message_from_local_queue)
                thread.daemon = True
                thread.start()
                self._message_handling_threads.append(thread)
        for _ in range(builder.get_number_of_amqp_consumers()):
            receive_callback = MessageSubscriberReceiveCallback(self)
            self._amqp_consumers.append(self._create_consumer(builder, receive_callback))
        if builder.get_register_default_binding():
            params = []
            params.append(self._queue_name)
            params.append('#')
            self.register_binding(params)
        if builder.get_auto_start_consumers():
            self.start_handling_messages()

    @classmethod
    def from_queue_name(cls, queue_name):
        """
        Create an instance of :class:`MessageSubscriber` based on the
        message queue name.

        :param queue_name: The message queue name.
        :type queue_name: str
        :return: An instance of :class:`MessageSubscriber`.
        :rtype: MessageSubscriber
        """
        builder = MessageSubscriberBuilder(queue_name)
        return cls(builder)

    def __del__(self):
        self.close()

    def close(self):
        """
        Close this instance of :class:`MessageSubscriber`.
        """
        if self._closing:
            return
        self._closing = True

        self.stop_handling_messages()
        if self._message_handling_threads:
            for _ in range(len(self._message_handling_threads)):
                self._local_message_queue.put(None)
            for thread in self._message_handling_threads:
                thread.join()
        if self._amqp_consumers is not None:
            for amqp_consumer in self._amqp_consumers:
                amqp_consumer.close()
            self._amqp_consumers = None

    def get_callback(self):
        """
        Get the callback object that is invoked whenever a message is
        received by this subscriber unless there is a more specific
        callback registered.

        :return: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :rtype: object
        """
        return self._callback

    def set_callback(self, callback):
        """
        Set the callback object that is invoked whenever a message is
        received by this subscriber unless there is a more specific
        callback registered.

        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._callback = callback

    def set_trace_logger(self, trace_logger):
        """
        Set the Trace Logger.

        :param trace_logger: The Trace Logger. May be ``None``
        :type trace_logger: niskyline.messagebus.trace_logger.TraceLogger
            or None
        """
        self._trace_logger = trace_logger

    def start_handling_messages(self):
        """
        Start handling incoming messages to all associated
        consumers.
        """
        for amqp_consumer in self._amqp_consumers:
            amqp_consumer.start_handling_messages()

    def stop_handling_messages(self):
        """
        Stop handling incoming messages to all associated
        consumers.
        """
        for amqp_consumer in self._amqp_consumers:
            amqp_consumer.stop_handling_messages()

    def acknowledge_message(self, consumer_tag, delivery_tag):
        """
        Acknowledge a message (meaning that it has been processed by a
        consumer that is part of this Message Subscriber).

        :param consumer_tag: The consumer tag. This is a unique identifier
            for a consumer.
        :type consumer_tag: str
        :param delivery_tag: The message delivery tag.
        :type delivery_tag: str
        """
        for consumer in self._amqp_consumers:
            if consumer.get_consumer_tag() == consumer_tag:
                consumer.acknowledge_message(delivery_tag)
                break

    def not_acknowledge_message(self, consumer_tag, delivery_tag):
        """
        Not acknowledge a message (meaning that it will not be processed by
        tha consumer that is part of this Message Subscriber).

        :param consumer_tag: The consumer tag. This is a unique identifier
            for a consumer.
        :type consumer_tag: str
        :param delivery_tag: The message delivery tag.
        :type delivery_tag: str
        """
        for consumer in self._amqp_consumers:
            if consumer.get_consumer_tag() == consumer_tag:
                consumer.not_acknowledge_message(delivery_tag)
                break

    @staticmethod
    def wait_for_message(queue_, block, timeout_seconds):
        """
        Wait for a message to appear on the queue.

        :param queue_: The queue to wait for the message on.
        :type queue_: queue.Queue
        :param block: ``True`` if we block to wait for the next
            message. ``False`` if we return immediately.
        :type block: bool
        :param timeout_seconds: The number of seconds to wait if ``block`` is
            ``True``. Use ``None`` to wait indefinitely. Ignored if ``block``
            is ``False``.
        :type timeout_seconds: int or float or None
        :return: The message or ``None`` if no message was found
            during the waiting period.
        :rtype: niskyline.messagebus.generic_message.GenericMessage or None
        """
        try:
            return queue_.get(block, timeout_seconds)
        except Empty:
            return None

    def register_message_name_callback(self, message_name, callback, explicit_ack=False):
        """
        Register a callback that will be invoked when a
        certain message name is used in an incoming message.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        :param explicit_ack: If ``True``, it means that the ``callback`` is
            responsible for acknowledging the message. If ``False``, it means
            that this Message Subscriber is responsible for acknowledging the
            message (if the
            :class:`niskyline.messagebus.message_subsciber_builder.MessageSubscriberBuilder`
            object used to create this object has enabled explicit
            acknowledgements).
        :type explicit_ack: bool
        """
        info = MessageNameCallbackInfo(callback, explicit_ack)
        self._msg_name_callbacks[message_name] = info

    def unregister_message_name_callback(self, message_name):  # pylint: disable=invalid-name
        """
        Unregister a callback that would have been invoked when a
        certain message name is used in an incoming message.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        """
        if message_name not in self._msg_name_callbacks:
            if self._trace_logger is not None:
                self._trace_logger.log_error(
                    'Failed to remove message name callback | ' +
                    self._queue_name +
                    ' | ' +
                    message_name
                )
        else:
            del self._msg_name_callbacks[message_name]

    def register_broadcast_callback(self, message_name, callback):
        """
        Register a callback that will be invoked when a
        certain message name is used in a broadcast message.

        :param message_name: The name of the message to use as the
            trigger for the callback.
        :type message_name: str
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        if message_name is None:
            return
        self.register_message_name_callback(message_name, callback)
        params = []
        params.append(message_name)
        self.register_binding(params)

    def register_correlation_id_callback(self, correlation_id, callback, state):  # pylint: disable=invalid-name
        """
        Register a callback that will be invoked when a
        certain correlation ID is used in an incoming message.

        :param correlation_id: The correlation ID to use as the
            trigger for the callback.
        :type correlation_id: str
        :param callback: A class instance that contains a method
            named ``callback``. The ``callback`` function takes an argument of
            type :class:`niskyline.messagebus.generic_message.GenericMessage`
            and a second argument that could be any type for user specific
            data.
        :type callback: object
        :param state: Optional user specific data. May be any
            type.
        :type state: object or None
        """
        if correlation_id is None:
            return
        info = CorrIdCallbackInfo(callback, state)
        self._corr_id_callbacks[correlation_id] = info

    def register_binding(self, binding_parameters):
        """
        Register a binding for all consumers.

        :param binding_parameters: A list of strings that represent a routing key.
        :type binding_parameters: list(str)
        """

        for consumer in self._amqp_consumers:
            consumer.register_binding(binding_parameters)

    def unregister_binding(self, binding_parameters):
        """
        Unregister a binding for all consumers.

        :param binding_parameters: A list of strings that represent a routing key.
        :type binding_parameters: list(str)
        """
        for consumer in self._amqp_consumers:
            consumer.unregister_binding(binding_parameters)

    def _process_message_from_local_queue(self):  # pylint: disable=invalid-name
        """
        Thread function for the local queue task threads.

        Will get a message from the local queue and process it.
        """
        try:
            while True:
                message = self._local_message_queue.get()
                if message is None:
                    break
                self._process_message(message)
                self._local_message_queue.task_done()
        except (KeyboardInterrupt, SystemExit):
            LOGGER.debug('MessageSubscriber _process_message_from_local_queue '
                         'exiting due to process exit')

    def _process_message(self, message):
        """
        Used to allow invoking user defined callbacks for incoming messages
        at various levels.

        Only one is invoked per message, in the following priority:
            1) Callbacks with ``state`` specific to the corrlation ID.
            2) Callbacks without ``state`` specific to the message name.
            3) General callbacks without ``state`` (user defined data).

        :param message: The incoming message to process.
        :type message: niskyline.messagebus.generic_message.GenericMessage
        """
        msg_name_callback_info = None
        corr_id_info = None
        if message.get_correlation_id() in self._corr_id_callbacks:
            corr_id_info = self._corr_id_callbacks[message.get_correlation_id()]
            del self._corr_id_callbacks[message.get_correlation_id()]
        if message.get_message_name() in self._msg_name_callbacks:
            msg_name_callback_info = self._msg_name_callbacks[message.get_message_name()]
        if corr_id_info is not None:
            corr_id_info.get_callback().callback(message, corr_id_info._state)  # pylint: disable=protected-access
        elif msg_name_callback_info is not None:
            if self._explicit_ack and not msg_name_callback_info.get_explicit_ack():
                self.acknowledge_message(message.get_consumer_tag(), message.GetDeliveryTag())
            if msg_name_callback_info.get_callback() is not None:
                msg_name_callback_info.get_callback().callback(message)
        elif self._callback is not None:
            self._callback.callback(message)
        elif message.has_error() and self._trace_logger is not None:
            self._trace_logger.log_error(
                'Unhandled message ' +
                message.get_message_name() +
                ' Origin=' + message.get_origin() +
                ' Error=' + message.get_error().to_string()
            )
        elif self._trace_logger is not None and self._trace_unhandled_message.is_enabled():
            self._trace_logger.Log(self._trace_unhandled_message, message.get_message_name())
        else:
            # TODO add support for creating a unhandled message callback.
            msg = 'Unhandled message: ' + message.get_message_name()
            if self._amqp_connection_manager is not None:
                self._amqp_connection_manager.put_status_message(
                    msg, False, replace_if_full=True
                )
            else:
                AmqpConnectionManager.get_instance().put_status_message(
                    msg, False, replace_if_full=True
                )
        self._purge_correlation_id_callbacks()

    def _purge_correlation_id_callbacks(self):  # pylint: disable=no-self-use
        """
        Traverse all stored correlation ID callbacks and remove them if
        they are older than the correlation ID timeout.
        """
        if USE_MONOTONIC_CLOCK:
            now = time.monotonic()  # pylint: disable=no-member
            diff_seconds = (
                (now - self._correlation_id_timeout) /
                self._clock_res
            )
            if diff_seconds < self._correlation_id_timeout:
                return
            self._correlation_id_timeout = now
            remove_list = []
            for corr_id, callback_info in self._corr_id_callbacks.items():
                diff_seconds = (
                    (now - callback_info.get_timestamp()) /
                    self._clock_res
                )
                if diff_seconds < self._correlation_id_timeout:
                    continue
                remove_list.append(corr_id)
        else:
            now = datetime.datetime.utcnow()
            diff_obj = now - self._correlation_id_timeout
            diff_seconds = diff_obj.microseconds() / 1000000.0
            if diff_seconds < self._correlation_id_timeout:
                return
            self._correlation_id_timeout = now
            remove_list = []
            for corr_id, callback_info in self._corr_id_callbacks.items():
                diff_obj = now - callback_info.get_timestamp()
                diff_seconds = diff_obj.microseconds() / 1000000.0
                if diff_seconds < self._correlation_id_timeout:
                    continue
                remove_list.append(corr_id)

        for corr_id in remove_list:
            del self._corr_id_callbacks[corr_id]
            if (self._trace_logger and self._trace_unhandled_message and
                    self._trace_unhandled_message.is_enabled()):
                self._trace_logger.log(
                    'Purged CorrelationIdCallback for ' + corr_id,
                    self._trace_unhandled_message
                )

    @staticmethod
    def _create_consumer(builder, callback):
        """
        Create an instance of
        :class:`niskyline.messagebus.amqp_consumer.AmqpConsumer`.

        :param builder: A
            :class:`niskyline.messagebus.message_subscriber_builder.MessageSubscriberBuilder`
            object.
        :type builder: niskyline.messagebus.message_subscriber_builder.MessageSubscriberBuilder
        :param callback: The callback for when handling messages. This
            is an instance of :class:`MessageSubscriberReceiveCallback`.
        :type callback: MessageSubscriberReceiveCallback
        :return: A new instance of
            :class:`niskyline.messagebus.amqp_consumer.AmqpConsumer`.
        :rtype: niskyline.messagebus.amqp_consumer.AmqpConsumer
        """
        LOGGER.debug('MessageSubscriber CreateConsumer!')
        consumer_builder = AmqpConsumerBuilder(builder.get_queue_name(), callback)
        consumer_builder.set_connection_manager(builder.get_connection_manager())
        consumer_builder.set_durable_queue(builder.get_durable_queue())
        consumer_builder.set_explicit_ack(builder.get_explicit_ack())
        consumer_builder.set_prefetch_queue_depth(builder.get_prefetch_queue_depth())
        consumer_builder.set_trace_logger(builder.get_trace_logger())
        return AmqpConsumer(consumer_builder)


class MessageNameCallbackInfo(object):
    """
    Data that is used for each registered callback that is based on
    message name.

    Used by :class:`MessageSubscriber`.
    """
    def __init__(self, callback=None, explicit_ack=None):
        """
        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        :param explicit_ack: If ``True``, it means that the ``callback`` is
            responsible for acknowledging the message. If ``False``, it means
            that this Message Subscriber is responsible for acknowledging the
            message (if the
            :class:`niskyline.messagebus.message_subsciber_builder.MessageSubscriberBuilder`
            object used to create this object has enabled explicit
            acknowledgements).
        :type explicit_ack: bool
        """
        self._callback = callback
        self._explicit_ack = explicit_ack

    def get_callback(self):
        """
        Get the callback object.

        :return: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :rtype: object
        """
        return self._callback

    def set_callback(self, callback):
        """
        Set the callback object.

        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._callback = callback

    def get_explicit_ack(self):
        """
        Get the explicit acknowledge mode.

        :return: If ``True``, it means that the ``callback`` is
            responsible for acknowledging the message. If ``False``, it means
            that this Message Subscriber is responsible for acknowledging the
            message (if the
            :class:`niskyline.messagebus.message_subsciber_builder.MessageSubscriberBuilder`
            object used to create this object has enabled explicit
            acknowledgements).
        :rtype: bool
        """
        return self._explicit_ack

    def set_explicit_ack(self, explicit_ack):
        """
        Set the explicit acknowledge mode.

        :param explicit_ack: If ``True``, it means that the ``callback`` is
            responsible for acknowledging the message. If ``False``, it means
            that this Message Subscriber is responsible for acknowledging the
            message (if the
            :class:`niskyline.messagebus.message_subsciber_builder.MessageSubscriberBuilder`
            object used to create this object has enabled explicit
            acknowledgements).
        :type explicit_ack: bool
        """
        self._explicit_ack = explicit_ack


class CorrIdCallbackInfo(object):
    """
    Data that is used for each registered callback that is based on
    correlation ID.

    Used by :class:`MessageSubscriber`.
    """
    def __init__(self, callback=None, state=None):
        """
        :param callback: A class instance that contains a method
            named ``callback``. The ``callback`` function takes an argument of
            type :class:`niskyline.messagebus.generic_message.GenericMessage`
            and a second argument that could be any type for user specific
            data. May by ``None``.
        :type callback: object or None
        :param state: Optional user specific data. May be any
            type.
        :type state: object or None
        """
        self._callback = callback
        self._state = state
        if USE_MONOTONIC_CLOCK:
            self._timestamp = time.monotonic()  # pylint: disable=no-member
        else:
            self._timestamp = datetime.datetime.utcnow()

    def get_callback(self):
        """
        Get the callback object.

        :return: A class instance that contains a method
            named ``callback``. The ``callback`` function takes an argument of
            type :class:`niskyline.messagebus.generic_message.GenericMessage`
            and a second argument that could be any type for user specific
            data. May by ``None``.
        :rtype: object or None
        """
        return self._callback

    def set_callback(self, callback):
        """
        Set the callback object.

        :param callback: A class instance that contains a method
            named ``callback``. The ``callback`` function takes an argument of
            type :class:`niskyline.messagebus.generic_message.GenericMessage`
            and a second argument that could be any type for user specific
            data. May by ``None``.
        :type callback: object or None
        """
        self._callback = callback

    def get_state(self):
        """
        Get the ``state`` (optional user specific data).

        :return: Optional user specific data. May be any
            type.
        :rtype: object or None
        """
        return self._state

    def set_state(self, state):
        """
        Set the ``state`` (optional user specific data).

        :param state: Optional user specific data. May be any
            type.
        :type state: object or None
        """
        self._state = state

    def get_timestamp(self):
        """
        Get the timestamp for this object. It is initially set when
        this object is created.

        :return: The timestamp for this object. It is based on an internal
            monotonic clock value.
        :rtype: float
        """
        return self._timestamp

    def set_timestamp(self, timestamp):
        """
        Set the timestamp for this object.

        :param timestamp: The timestamp for this object. It is based on an internal
            monotonic clock value.
        :type timestamp: float
        """
        self._timestamp = timestamp


class SynchCallbackInfo(object):
    """
    Data that is used when a synchronous query is used.

    Used by :class:`niskyline.messagebus.message_service.MessageService`.
    """
    def __init__(self, queue_=None):
        """
        :param queue_: The queue to use for synchronous callbacks. May be
            ``None``.
        :type queue_: queue.Queue or None
        """
        self._queue = queue_

    def get_queue(self):
        """
        Get the queue to use for synchronous callbacks.

        :return: The queue to use for synchronous callbacks. May be
            ``None``.
        :rtype: queue.Queue or None
        """
        return self._queue

    def set_queue(self, queue_):
        """
        Set the queue to use for synchronous callbacks.

        :param queue_: The queue to use for synchronous callbacks. May be
            ``None``.
        :type queue_: queue.Queue or None
        """
        self._queue = queue_


class MessageSubscriberReceiveCallback(object):
    """
    Used for handling incoming messages.

    Used by :class:`MessageSubscriber`,
    :class:`niskyline.messagebus.amqp_consumer_builder.AmqpConsumerBuilder`,
    and :class:`niskyline.messagebus.amqp_consumer.AmqpConsumer`.
    """
    def __init__(self, message_subscriber):
        """
        param message_subscriber: An instance of :class:`MessageSubscriber`.
        type message_subscriber: MessageSubscriber
        """
        self._message_subscriber = message_subscriber

    def handle_message(self, generic_message):
        """
        Handle a message by either queuing it to let the local message
        queue task threads process it, or process it directly if such
        threads are not set up.

        :param generic_message: The incoming message to handle. The
            class of the message is
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type generic_message: niskyline.messagebus.generic_message.GenericMessage
        """
        info = self._message_subscriber._synch_callbacks.get(  # pylint: disable=protected-access
            generic_message.get_correlation_id()
        )
        if info is not None:
            info._queue.put(generic_message)  # pylint: disable=protected-access
            # info.reset_event.set()
        if self._message_subscriber._local_message_queue is not None:  # pylint: disable=protected-access
            self._message_subscriber._local_message_queue.put(generic_message)  # pylint: disable=protected-access
        else:
            self._message_subscriber._process_message(generic_message)  # pylint: disable=protected-access

    def handle_reset_event(self, reset_event):
        """
        Handle the case when a mandatory delivery could not be honored.

        We must cancel the synchronous wait (to avoid waiting for the
        rest of the timeout) and remove the synchronous and
        correlation ID callbacks.
        """
        info = self._message_subscriber._synch_callbacks.pop(  # pylint: disable=protected-access
            reset_event.get_correlation_id(), None
        )
        if info is not None:
            queue_ = info.get_queue()
            while not queue_.empty():
                queue_.get_nowait()
            # Cancel the synchronous callback that is waiting on this correlation id.
            queue_.put(reset_event.get_event_message())

        self._message_subscriber._corr_id_callbacks.pop(  # pylint: disable=protected-access
            reset_event.get_correlation_id(), None
        )
