# -*- coding: utf-8 -*-
"""
Implementation of 'MessageSubscriberBuilder' class
"""
from __future__ import absolute_import


class MessageSubscriberBuilder(object):  # pylint: disable=too-many-public-methods, too-many-instance-attributes
    """
    Message Subscriber Builder.

    This is used for the construction of a
    :class:`niskyline.messagebus.message_subscriber.MessageSubscriber` object.
    """
    def __init__(self, queue_name=None):
        self._queue_name = queue_name
        self._number_of_amqp_consumers = 1
        self._max_local_queue_depth = 100
        self._local_queue_task_threads = 1
        self._prefetch_queue_depth = 100
        self._auto_start_consumers = True
        self._register_default_binding = True
        # Default to 6 hours... is this reasonable?
        self._correlation_id_timeout = 6 * 60 * 60
        self._durable_queue = False
        self._explicit_ack = False
        self._trace_logger = None
        self._callback = None
        self._connection_manager = None

    def get_connection_manager(self):
        """
        Get the Connection Manager. If this is ``None``, a new
        Connection Manager will be created during the creation
        of :class:`niskyline.messagebus.message_subscriber.MessageSubscriber`.

        :return: The Connection Manager. May be ``None``.
        :rtype: niskyline.messagebus.amqp_connection_manager.AmqpConnectionManager
            or None
        """
        return self._connection_manager

    def set_connection_manager(self, connection_manager):
        """
        Set the Connection Manager. If this is ``None``, a new
        Connection Manager will be created during the creation
        of :class:`niskyline.messagebus.message_subscriber.MessageSubscriber`.

        :param connection_manager: The Connection Manager. May be ``None``.
        :type connection_manager:
            niskyline.messagebus.amqp_connection_manager.AmqpConnectionManager
            or None
        """
        self._connection_manager = connection_manager

    def get_queue_name(self):
        """
        Get the message queue name.

        :return: The message queue name.
        :rtype: str
        """
        return self._queue_name

    def set_queue_name(self, queue_name):
        """
        Set the message queue name.

        :param queue_name: The message queue name.
        :type queue_name: str
        """
        self._queue_name = queue_name

    def get_callback(self):
        """
        Get the callback object that is invoked whenever a message is
        received by the Message Subscriber unless there is a more specific
        callback registered.

        :return: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :rtype: object
        """
        return self._callback

    def set_callback(self, callback):
        """
        Set the callback object that is invoked whenever a message is
        received by the Message Subscriber unless there is a more specific
        callback registered.

        :param callback: A class instance that contains a method named ``callback``.
            The ``callback`` function takes an argument of type
            :class:`niskyline.messagebus.generic_message.GenericMessage`.
        :type callback: object
        """
        self._callback = callback

    def get_number_of_amqp_consumers(self):
        """
        Get the number of
        :class:`niskyline.messagebus.amqp_consumer.AmqpConsumer` objects
        that will be created.

        :return: The number of
            :class:`niskyline.messagebus.amqp_consumer.AmqpConsumer`
            objects that will be created.
        :rtype: int
        """
        return self._number_of_amqp_consumers

    def set_number_of_amqp_consumers(self, number_of_amqp_consumers):
        """
        Set the number of
        :class:`niskyline.messagebus.amqp_consumer.AmqpConsumer` objects
        that will be created.

        :param number_of_amqp_consumers: The number of
            :class:`niskyline.messagebus.amqp_consumer.AmqpConsumer`
            objects that will be created.
        :type number_of_amqp_consumers: int
        """
        self._number_of_amqp_consumers = number_of_amqp_consumers

    def get_max_local_queue_depth(self):
        """
        Get the maximum number of items that may be placed in the local
        message queue.

        :return: The maximum number of items that may be placed in the local
            message queue.
        :rtype: int
        """
        return self._max_local_queue_depth

    def set_max_local_queue_depth(self, max_local_queue_depth):
        """
        Set the maximum number of items that may be placed in the local
        message queue.

        :param max_local_queue_depth: The maximum number of items that
            may be placed in the local message queue.
        :type max_local_queue_depth: int
        """
        self._max_local_queue_depth = max_local_queue_depth

    def get_local_queue_task_threads(self):
        """
        Get the number of task threads to use when processing items in
        the local message queue.

        :return: The number of task threads to use when processing items in
            the local message queue.
        :rtype: int
        """
        return self._local_queue_task_threads

    def set_local_queue_task_threads(self, local_queue_task_threads):
        """
        Set the number of task threads to use when processing items in
        the local message queue.

        :param local_queue_task_threads: The number of task threads to
            use when processing items in the local message queue.
        :type local_queue_task_threads: int
        """
        self._local_queue_task_threads = local_queue_task_threads

    def get_durable_queue(self):
        """
        Get whether the queue is ``durable``. A ``durable`` queue can
        survive a broker restart, however, only ``persistent`` messages
        will be recovered.

        :return: ``True`` if the queue is ``durable``. ``False``
            otherwise.
        :rtype: bool
        """
        return self._durable_queue

    def set_durable_queue(self, durable_queue):
        """
        Set whether the queue is ``durable``. A ``durable`` queue can
        survive a broker restart, however, only ``persistent`` messages
        will be recovered.

        :param durable_queue: ``True`` if the queue is ``durable``.
            ``False`` otherwise.
        :type durable_queue: bool
        """
        self._durable_queue = durable_queue

    def get_explicit_ack(self):
        """
        Get whether the
        :class:`niskyline.messagebus.message_subscriber.MessageSubscriber`
        object is responsible for acknowledging incoming messages.

        :return: ``True`` if the
            :class:`niskyline.messagebus.message_subscriber.MessageSubscriber`
            object is responsible for acknowledging incoming messages.
            ``False`` otherwise.
        :rtype: bool
        """
        return self._explicit_ack

    def set_explicit_ack(self, explicit_ack):
        """
        Set whether the
        :class:`niskyline.messagebus.message_subscriber.MessageSubscriber`
        object is responsible for acknowledging incoming messages.

        :param explicit_ack: ``True`` if the
            :class:`niskyline.messagebus.message_subscriber.MessageSubscriber`
            object is responsible for acknowledging incoming messages.
            ``False`` otherwise.
        :type explicit_ack: bool
        """
        self._explicit_ack = explicit_ack

    def get_prefetch_queue_depth(self):
        """
        Get the prefetch queue depth.

        :return: The prefetch queue depth.
        :rtype: int
        """
        return self._prefetch_queue_depth

    def set_prefetch_queue_depth(self, prefetch_queue_depth):
        """
        Set the prefetch queue depth.

        :param prefetch_queue_depth: The prefetch queue depth.
        :type prefetch_queue_depth: int
        """
        self._prefetch_queue_depth = prefetch_queue_depth

    def get_trace_logger(self):
        """
        Get the Trace Logger.

        :return: The Trace Logger.
        :rtype: niskyline.messagebus.trace_logger.TraceLogger
        """
        return self._trace_logger

    def set_trace_logger(self, trace_logger):
        """
        Set the Trace Logger.

        :param trace_logger: The Trace Logger.
        :type trace_logger: niskyline.messagebus.trace_logger.TraceLogger
        """
        self._trace_logger = trace_logger

    def get_auto_start_consumers(self):
        """
        Get whether to auto start consumers.

        :return: ``True`` if the consumers will be started during
            creation of
            :class:`niskyline.messagebus.message_subscriber.MessageSubscriber`.
            ``False`` otherwise.
        :rtype: bool
        """
        return self._auto_start_consumers

    def set_auto_start_consumers(self, auto_start_consumers):
        """
        Set whether to auto start consumers.

        :param auto_start_consumers: ``True`` if the consumers will be
            started during creation of
            :class:`niskyline.messagebus.message_subscriber.MessageSubscriber`.
            ``False`` otherwise.
        :type auto_start_consumers: bool
        """
        self._auto_start_consumers = auto_start_consumers

    def get_register_default_binding(self):
        """
        Get whether to use the default registration binding. If so, it
        will automatically register the binding during creation of
        :class:`niskyline.messagebus.message_subscriber.MessageSubscriber`.

        :return: ``True`` to use the default registration binding.
            ``False`` otherwise.
        :rtype: bool
        """
        return self._register_default_binding

    def set_register_default_binding(self, register_default_binding):
        """
        Set whether to use the default registration binding. If so, it
        will automatically register the binding during creation of
        :class:`niskyline.messagebus.message_subscriber.MessageSubscriber`.

        :param register_default_binding: ``True`` to use the default
            registration binding. ``False`` otherwise.
        :type register_default_binding: bool
        """
        self._register_default_binding = register_default_binding

    def get_correlation_id_timeout(self):
        """
        Get the correlation ID timeout (in seconds).

        :return: The correlation ID timeout (in seconds).
        :rtype: float or int
        """
        return self._correlation_id_timeout

    def set_correlation_id_timeout(self, correlation_id_timeout):
        """
        Set the correlation ID timeout (in seconds).

        :param correlation_id_timeout: The correlation ID timeout
            (in seconds).
        :type correlation_id_timeout: float or int
        """
        self._correlation_id_timeout = correlation_id_timeout
