# -*- coding: utf-8 -*-
"""
Implementation of 'TraceLogger' class
"""
from __future__ import absolute_import

# Import local libs
# pylint: disable=import-error,no-name-in-module
from niskyline.messagebus.error import Error
# pylint: enable=import-error,no-name-in-module


class TraceLogger(object):
    """
    The Trace Logger.

    A class that logs various types of information to the Skyline log files.
    """
    def __init__(self, name, parent, process_logger):
        """
        :param name: The last part of the name to use for this Trace Logger.
            Will not be the full name if ``parent`` is not ``None.
            May be ``None``.
        :type name: str or None
        :param parent: The parent Trace Logger object used to create this one.
            May be ``None``.
        :type parent: TraceLogger
        :param process_logger: The Process Logger associated with this Trace
            Point.
        :type process_logger: niskyline.messagebus.process_logger.ProcessLogger
        """
        self._name = ''
        self._process_logger = None

        builder = ''
        if parent:
            builder += parent.get_name() + '.'
        if name:
            builder += name
        self.initialize(builder, process_logger)

    def initialize(self, name, process_logger):
        """
        Initialize this :class:`TraceLogger` object.

        :param name: The full name to use for this Trace Logger.
        :type name: str
        :param process_logger: The Process Logger associated with this Trace
            Point.
        :type process_logger: niskyline.messagebus.process_logger.ProcessLogger
        """
        self._name = name
        self._process_logger = process_logger

    def get_name(self):
        """
        Get the full name of this :class:`TraceLogger` object.

        :return: The full name of this :class:`TraceLogger` object.
        :rtype: str
        """
        return self._name

    def set_name(self, name):
        """
        Set the full name of this :class:`TraceLogger` object.

        :param name: The full name of this :class:`TraceLogger` object.
        :type name: str
        """
        self._name = name

    def make_trace_point(self, name):
        """
        Create a Trace Point.

        :param name: Name of the Trace Point.
        :type name: str
        :return: The associated Trace Point.
        :rtype: niskyline.messagebus.trace_point.TracePoint
        """
        builder = self._name + '.' + name
        return self._process_logger.make_or_lookup_trace_point(builder)

    def log(self, log_string, trace_point=None):
        """
        Log to the Trace Logger via a Trace Point.

        :param log_string: String to log.
        :type log_string: str
        :param trace_point: Trace point to use or ``None`` to not use a
            trace point.
        :type trace_point: niskyline.messagebus.trace_point.TracePoint
        """
        if trace_point is None:
            self.log_error('Null tracepoint: ' + log_string)
        if trace_point.is_enabled():
            self._process_logger.log(self._name, trace_point, log_string)

    def log_info(self, log_string):
        """
        Log an informative message to the Trace Logger.

        :param log_string: String to log.
        :type log_string: str
        """
        self._process_logger.log_info(self._name, log_string)

    def log_error(self, error):
        """
        Log an error to the Trace Logger.

        :param error: The error may be a :class:`str`, a
            :class:`niskyline.messagebus.error.Error` object, or or a list of
            :class:`niskyline.messagebus.error.Error` objects.
        :type error: str or niskyline.messagebus.error.Error or
            list(niskyline.messagebus.error.Error)
        """
        if isinstance(error, str):
            self._process_logger.log_error(self._name, error)
        elif isinstance(error, Error):
            self._process_logger.log_error(self._name, error.to_string())
        elif isinstance(error, list):
            for error_item in error:
                self._process_logger.log_error(
                    self._name, error_item.to_string()
                )

    def perform_assert(self, assertion, log_string):
        """
        Test an assertion and log an error to the Trace Logger if it fails.

        :param assertion: Expression to evaluate as an assertion.
        :type assertion: bool
        :param log_string: Error message to log if the assertion fails.
        :type log_string: str
        """
        if not assertion:
            self._process_logger.log_error(self._name, log_string)
