# -*- coding: utf-8 -*-
"""
API to create, update, and query Skyline Tags.
"""
from __future__ import absolute_import

# Import python libs
import logging
import threading

# Import local libs
# pylint: disable=import-error
from niskyline.messagebus.managed_service_base import ManagedServiceBase
from niskyline.messagebus.managed_service_builder import ManagedServiceBuilder
from niskyline.messagebus.message_header import MessageHeader
from . import ping_pong_messages
# pylint: enable=import-error

# Set up logging
LOGGER = logging.getLogger(__name__)


class PingPongService(ManagedServiceBase):
    """
    Class to publicly access the Ping Pong Service.
    """
    def __init__(self, service_name, standalone=False):
        """
        :param service_name: The name of the service to use within
            the Skyline framework.
        :type service_name: str
        """
        builder = ManagedServiceBuilder(service_name)
        builder.set_standalone_property(standalone)
        shutdown_event = threading.Semaphore(0)
        super(PingPongService, self).__init__(
            service_name=None,
            shutdown_event=shutdown_event,
            managed_service_builder=builder)
        self.register_work_subscriber_callback(
            ping_pong_messages.PingPongConcatRequest.__name__,
            PingPongConcatRequestCallback(self))
        self.initialize()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def close(self):  # pylint: disable=useless-super-delegation
        """
        Close the PingPongService and all associated resources.
        """
        super(PingPongService, self).close()

    def get_shutdown_event(self):
        """
        Get the shutdown event.

        :return: A :class`threading.Semaphore` instance
            that is released when the Skyline Service Manager requires
            the service to shut down.
        :rtype: threading.Semaphore
        """
        return self._shutdown_event


class PingPongConcatRequestCallback(object):  # pylint: disable=too-few-public-methods
    """
    Class to handle incoming request messages to the Ping Pong Service.
    """
    def __init__(self, managed_service):
        """
        :param managed_service: An instance of the PingPongService.
        :type managed_service: PingPongService
        """
        self._managed_service = managed_service

    def callback(self, generic_message):
        """
        A callback from the Skyline Message Bus when there
        is an incoming message to handle.

        :param generic_message: An incoming message to handle.
        :type generic_message: niskyline.messagebus.generic_message.GenericMessage
        """
        LOGGER.debug('Received PingPongConcatRequest!')
        ping_pong_concat_request = ping_pong_messages.PingPongConcatRequest.from_message(
            generic_message
        )
        LOGGER.debug('a = %s', ping_pong_concat_request.a)
        LOGGER.debug('b = %s', ping_pong_concat_request.b)
        result = ping_pong_concat_request.a + ping_pong_concat_request.b
        ping_pong_concat_response = ping_pong_messages.PingPongConcatResponse(
            request_message=ping_pong_concat_request,
            result_=result)
        self._managed_service.publish_response(ping_pong_concat_response)
