# -*- coding: utf-8 -*-
"""
Command line actions for the Ping Pong Service.
"""
from __future__ import absolute_import, print_function

# Import python libs
import argparse
import json
import logging
import os.path
import subprocess
import sys
import time

# Import third party libs
import psutil

# Import local libs
# pylint: disable=import-error
import niskyline.messagebus.utils as utils
import niskyline.messagebus.service_manager as service_manager
from niskyline.pingpongservice import PingPongService
from niskyline.messagebus.paths import get_service_descriptor_install_directory
# pylint: enable=import-error


LOG_LEVELS = {
    'all': logging.NOTSET,
    'debug': logging.DEBUG,
    'error': logging.ERROR,
    'critical': logging.CRITICAL,
    'info': logging.INFO,
    'warning': logging.WARNING
}

SORTED_LEVEL_NAMES = [
    log_item[0] for log_item in sorted(LOG_LEVELS.items(), key=lambda x: x[1])
]

DEFAULT_LOG_LEVEL = 'warning'

COMMAND_CHOICES = [
    'sync', 'start', 'stop', 'status', 'sync-standalone',
    'start-standalone', 'stop-standalone',
    'status-standalone', 'register', 'unregister'
]

PING_PONG_SERVICE_DESC = 'PingPong.json'
SERVICE_NAME = 'PingPong'

DETACHED_PROCESS = 0x00000008

ARGUMENTS_TOKEN = '\"Arguments\":'
PATH_TOKEN = '\"Path\":'
PYTHONPATH_TOKEN = '\"PYTHONPATH\":'


def parse_args():
    """
    This will parse the command-line arguments passed in.
    Returns a Namespace object of the parsed command-line arguments.

    :return: A Namespace object of the parsed command-line arguments.
    :rtype: argparse.Namespace
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-l', '--log-level',
        dest='log_level',
        choices=SORTED_LEVEL_NAMES,
        default=DEFAULT_LOG_LEVEL,
        help='Console logging log level. One of {0}. Default: \'{1}\'.'.format(
            ', '.join([repr(l) for l in SORTED_LEVEL_NAMES]),
            DEFAULT_LOG_LEVEL)
    )
    parser.add_argument(
        '-s', '--silent',
        dest='silent',
        action='store_true',
        help='Silent mode'
    )
    parser.add_argument(
        'command',
        choices=COMMAND_CHOICES,
        help='Run a ping pong command. One of {0}.'.format(
            ', '.join([repr(cmd) for cmd in COMMAND_CHOICES]))
    )
    parser.add_argument(
        'command-args',
        nargs=argparse.REMAINDER,
        help='Arguments for a specific command.'
    )
    args = parser.parse_args()
    return args


def setup_logging(args):
    """
    This will set up logging at the desired log level.

    :param args: A Namespace object of the parsed command-line arguments.
    :type args: argparse.Namespace
    """
    log_level = LOG_LEVELS[args.log_level]
    log = logging.getLogger()
    log.setLevel(log_level)

    # Create console handler and set level to debug
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(log_level)

    # Create formatter
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # Add formatter to stream_handler
    stream_handler.setFormatter(formatter)

    # Add handler to logger
    log.addHandler(stream_handler)


def run_service(standalone, silent=False):
    """
    This will run the service.

    :param standalone: If ``True``, run in standalone mode. If ``False``,
        must be launched from within the Service Manager.
    :type standalone: bool
    :param silent: If ``False``, will print status to stdout/stderr.
        If ``True``, will not print status.
    :type silent: bool
    :return: The return code. ``0`` on success. Non-zero on failure.
    :rtype: int
    """
    try:
        with PingPongService(service_name=SERVICE_NAME, standalone=standalone) as service:
            shutdown_event = service.get_shutdown_event()
            if silent:
                shutdown_event.acquire(True)
            else:
                print('Waiting for messages...', end='')
                while True:
                    time.sleep(1)
                    print('.', end='', flush=True)
                    if shutdown_event.acquire(False):
                        break
    except (KeyboardInterrupt, SystemExit):
        if not silent:
            print('Ping Pong Service has stopped.')
    return 0


def is_standalone_service_running(terminate=False):
    """
    This will return ``True`` if the service is already running.

    :param terminate: If ``True``, will terminate the service
        if it is already running.
    :type terminate: bool
    :return: ``True`` if the service is already running.
        ``False`` otherwise.
    :rtype: bool
    """
    abs_filename = os.path.abspath(__file__)
    for proc in psutil.process_iter():
        try:
            cmdline = proc.cmdline()
            if (abs_filename in cmdline and
                    '--silent' in cmdline and
                    'sync-standalone' in cmdline):
                if terminate:
                    try:
                        proc.terminate()
                    except psutil.AccessDenied:
                        msg = 'Insufficient privileges to stop the service'
                        utils.eprint(msg)
                        if sys.version_info[0] == 2:
                            raise OSError(msg)
                        else:
                            raise PermissionError(msg)  # pylint: disable=undefined-variable
                return True
        except (psutil.AccessDenied, psutil.ZombieProcess,
                psutil.NoSuchProcess):
            pass
    return False


def start_service(silent=False):
    """
    This will start the service.

    :param silent: If ``False``, will print status to stdout/stderr.
        If ``True``, will not print status.
    :type silent: bool
    :return: The return code. ``0`` on success. Non-zero on failure.
    :rtype: int
    """
    if is_service_running():
        if not silent:
            print('Ping Pong Service already running')
        return 0
    service_manager.start_service(SERVICE_NAME)
    if not silent:
        print('Ping Pong Service successfully started')
    return 0


def stop_service(silent=False):
    """
    This will stop the service.

    :param silent: If ``False``, will print status to stdout/stderr.
        If ``True``, will not print status.
    :type silent: bool
    :return: The return code. ``0`` on success. Non-zero on failure.
    :rtype: int
    """
    service_manager.stop_service(SERVICE_NAME)
    if not silent:
        print('Ping Pong Service stopped')
    return 0


def is_service_running():
    """
    This will return ``True`` if the service is already running.

    :return: ``True`` if the service is already running.
        ``False`` otherwise.
    :rtype: bool
    """
    services = service_manager.get_services()
    for service in services:
        if (service.name == SERVICE_NAME and
                service.service_state.to_string() == 'LIVE'):
            return True
    return False


def start_standalone_service(silent=False):
    """
    This will start the service.

    :param silent: If ``False``, will print status to stdout/stderr.
        If ``True``, will not print status.
    :type silent: bool
    :return: The return code. ``0`` on success. Non-zero on failure.
    :rtype: int
    """
    if is_standalone_service_running():
        if not silent:
            print('Ping Pong Service already running')
        return 0

    subprocess.Popen(
        [sys.executable, os.path.abspath(__file__), '--silent', 'sync-standalone'],
        creationflags=DETACHED_PROCESS
    )
    if not silent:
        print('Ping Pong Service successfully started')
    return 0


def stop_standalone_service(silent=False):
    """
    This will stop the service.

    :param silent: If ``False``, will print status to stdout/stderr.
        If ``True``, will not print status.
    :type silent: bool
    :return: The return code. ``0`` on success. Non-zero on failure.
    :rtype: int
    """
    if is_standalone_service_running(terminate=True):
        if not silent:
            print('Ping Pong Service stopped')
    else:
        if not silent:
            print('Ping Pong Service is not running')
    return 0


def register_service(silent=False):
    """
    This will register this service with the NI Skyline Service Manager.

    :param silent: If ``False``, will print status to stdout/stderr.
        If ``True``, will not print status.
    :type silent: bool
    :return: The return code. ``0`` on success. Non-zero on failure.
    :rtype: int
    """
    if not utils.is_windows():
        utils.eprint('This operation is currently only supported in Windows')
        return 1

    source_desc = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), 'install',
        PING_PONG_SERVICE_DESC
    )

    dest_desc = os.path.join(
        get_service_descriptor_install_directory(),
        PING_PONG_SERVICE_DESC
    )

    if os.path.isfile(dest_desc):
        if not silent:
            print('Ping Pong Service already registered')
        return 0

    python_path = os.environ.get('PYTHONPATH')
    if not python_path:
        python_path = ''
    script_cmd = '\"{0}\" --silent sync'.format(__file__)

    with open(source_desc, 'r', encoding='utf-8') as read_fp:
        desc_data = json.loads(read_fp.read())

    service_data = desc_data['ServiceDescriptor']['ServiceGroups'][0]
    service_data['Properties']['Path'] = sys.executable
    service_data['Properties']['Arguments'] = script_cmd
    service_data['EnvironmentVariables']['PYTHONPATH'] = python_path

    file_contents = json.dumps(desc_data, indent=2)

    with open(dest_desc, 'w', encoding='utf-8') as write_fp:
        write_fp.write(file_contents)

    service_manager.restart_service_manager(silent=silent)
    if not silent:
        print('Ping Pong Service successfully registered')
    return 0


def unregister_service(silent=False):
    """
    This will unregister this service with the NI Skyline Service Manager.

    :param silent: If ``False``, will print status to stdout/stderr.
        If ``True``, will not print status.
    :type silent: bool
    :return: The return code. ``0`` on success. Non-zero on failure.
    :rtype: int
    """
    if not utils.is_windows():
        utils.eprint('This operation is currently only supported in Windows')
        return 1

    dest_desc = os.path.join(
        get_service_descriptor_install_directory(),
        PING_PONG_SERVICE_DESC
    )

    retcode = 0
    if os.path.isfile(dest_desc):
        os.remove(dest_desc)
        service_manager.restart_service_manager(silent=silent)
        if not silent:
            print('Ping Pong Service successfully unregistered')
    else:
        if not silent:
            print('Ping Pong Service already not registered')
    return retcode


def main():
    """
    This is the main entry point for this script.
    """
    args = parse_args()
    setup_logging(args)

    command = args.command
    command_args = getattr(args, 'command-args', [])  # pylint: disable=unused-variable

    retcode = 0
    # parse_args() will ensure that the command is a valid option.
    try:
        if command == 'sync':
            retcode = run_service(False, silent=args.silent)
        elif command == 'start':
            retcode = start_service(silent=args.silent)
        elif command == 'stop':
            retcode = stop_service(silent=args.silent)
        elif command == 'status':
            if is_service_running():
                if not args.silent:
                    print('Ping Pong Service is running')
            else:
                if not args.silent:
                    print('Ping Pong Service is not running')
                retcode = 1
        elif command == 'sync-standalone':
            retcode = run_service(True, silent=args.silent)
        elif command == 'start-standalone':
            retcode = start_standalone_service(silent=args.silent)
        elif command == 'stop-standalone':
            retcode = stop_standalone_service(silent=args.silent)
        elif command == 'status-standlone':
            if is_standalone_service_running():
                if not args.silent:
                    print('Ping Pong Service is running')
            else:
                if not args.silent:
                    print('Ping Pong Service is not running')
                retcode = 1
        elif command == 'register':
            retcode = register_service(silent=args.silent)
        elif command == 'unregister':
            retcode = unregister_service(silent=args.silent)
    except Exception:  # pylint: disable=broad-except
        retcode = 2
        if not args.silent:
            # Raising the exception here will print exception info to
            # the console.
            raise

    if retcode != 0:
        sys.exit(retcode)


if __name__ == '__main__':
    main()
